"""PDF lexer and parser."""

import logging
import mmap
import re
from binascii import unhexlify
from collections import deque
from typing import (
    TYPE_CHECKING,
    Any,
    Deque,
    Dict,
    Final,
    Iterable,
    Iterator,
    List,
    NamedTuple,
    Tuple,
    Union,
)

from playa.exceptions import PDFSyntaxError
from playa.pdftypes import (
    KWD,
    LIT,
    LITERALS_ASCII85_DECODE,
    LITERALS_ASCIIHEX_DECODE,
    ContentStream,
    InlineImage,
    ObjRef,
    PDFObject,
    PSKeyword,
    PSLiteral,
    decipher_all,
    int_value,
    literal_name,
    name_str,
    stream_value,
)
from playa.utils import choplist
from playa.worker import _deref_document, _ref_document

if TYPE_CHECKING:
    from playa.document import Document

log: Final = logging.getLogger(__name__)
# Intern a bunch of important keywords
KEYWORD_PROC_BEGIN: Final = KWD(b"{")
KEYWORD_PROC_END: Final = KWD(b"}")
KEYWORD_ARRAY_BEGIN: Final = KWD(b"[")
KEYWORD_ARRAY_END: Final = KWD(b"]")
KEYWORD_DICT_BEGIN: Final = KWD(b"<<")
KEYWORD_DICT_END: Final = KWD(b">>")
KEYWORD_GT: Final = KWD(b">")
KEYWORD_R: Final = KWD(b"R")
KEYWORD_NULL: Final = KWD(b"null")
KEYWORD_ENDOBJ: Final = KWD(b"endobj")
KEYWORD_STREAM: Final = KWD(b"stream")
KEYWORD_ENDSTREAM: Final = KWD(b"endstream")
KEYWORD_XREF: Final = KWD(b"xref")
KEYWORD_STARTXREF: Final = KWD(b"startxref")
KEYWORD_OBJ: Final = KWD(b"obj")
KEYWORD_TRAILER: Final = KWD(b"trailer")
KEYWORD_BI: Final = KWD(b"BI")
KEYWORD_ID: Final = KWD(b"ID")
KEYWORD_EI: Final = KWD(b"EI")


EOL: Final = b"\r\n"
WHITESPACE: Final = b" \t\n\r\f\v"
NUMBER: Final = b"0123456789"
HEX: Final = NUMBER + b"abcdef" + b"ABCDEF"
NOTLITERAL: Final = b"#/%[]()<>{}" + WHITESPACE
NOTKEYWORD: Final = b"#/%[]()<>{}" + WHITESPACE
NOTSTRING: Final = b"()\\"
OCTAL: Final = b"01234567"
ESC_STRING: Final = {
    ord(b"b"): b"\b",
    ord(b"t"): b"\t",
    ord(b"n"): b"\n",
    ord(b"f"): b"\f",
    ord(b"r"): b"\r",
}


Token = Union[float, bool, PSLiteral, PSKeyword, bytes]
LEXER: Final = re.compile(
    rb"""(?:
      (?P<whitespace> \s+)
    | (?P<comment> %[^\r\n]*[\r\n])
    | (?P<name> /(?: \#[A-Fa-f\d][A-Fa-f\d] | [^#/%\[\]()<>{}\s])* )
    | (?P<number> [+-]? (?: \. \d+ | \d+ (?: \. \d* )?))
    | (?P<keyword> [A-Za-z] [^#/%\[\]()<>{}\s]*)
    | (?P<startstr> \([^()\\]*)
    | (?P<hexstr> <[A-Fa-f\d\s]*>)
    | (?P<startdict> <<)
    | (?P<enddict> >>)
    | (?P<other> .)
)
""",
    re.VERBOSE,
)
STRLEXER: Final = re.compile(
    rb"""(?:
      (?P<octal> \\[0-7]{1,3})
    | (?P<linebreak> \\(?:\r\n?|\n))
    | (?P<escape> \\.)
    | (?P<parenleft> \()
    | (?P<parenright> \))
    | (?P<newline> \r\n?|\n)
    | (?P<other> .)
)""",
    re.VERBOSE,
)
HEXDIGIT: Final = re.compile(rb"#([A-Fa-f\d][A-Fa-f\d])")
EOLR: Final = re.compile(rb"\r\n?|\n")
SPC: Final = re.compile(rb"\s")
WSR: Final = re.compile(rb"\s+")
DOT: Final[int] = ord(b".")


class Lexer(Iterator[Tuple[int, Token]]):
    """Lexer for PDF data."""

    def __init__(self, data: Union[bytes, mmap.mmap], pos: int = 0) -> None:
        self.data = data
        self.pos = pos
        self.end = len(data)
        self._tokens: Deque[Tuple[int, Token]] = deque()
        self._curtoken: bytes = b""
        self._curtokenpos: int = 0

    def seek(self, pos: int) -> None:
        """Seek to a position and reinitialize parser state."""
        self.pos = pos
        self._curtoken = b""
        self._curtokenpos = 0
        self._tokens.clear()

    def tell(self) -> int:
        """Get the current position in the buffer."""
        return self.pos

    def read(self, objlen: int) -> bytes:
        """Read data from current position, advancing to the end of
        this data."""
        pos = self.pos
        self.pos = min(pos + objlen, len(self.data))
        return self.data[pos : self.pos]

    def nextline(self) -> Tuple[int, bytes]:
        r"""Get the next line ending either with \r, \n, or \r\n,
        starting at the current position."""
        linepos = self.pos
        m = EOLR.search(self.data, self.pos)
        if m is None:
            self.pos = self.end
        else:
            self.pos = m.end()
        return (linepos, self.data[linepos : self.pos])

    def __next__(self) -> Any:  # should be Tuple[int, Token] but mypyc
        """Get the next token in iteration, raising StopIteration when
        done."""
        while True:
            m = LEXER.match(self.data, self.pos)
            if m is None:  # can only happen at EOS
                raise StopIteration
            self._curtokenpos = self.pos
            self.pos = m.end()
            if m.lastgroup not in ("whitespace", "comment"):
                # Okay, we got a token or something
                break
        self._curtoken = m[0]
        if m.lastgroup == "number":
            if DOT in self._curtoken:
                return (self._curtokenpos, float(self._curtoken))
            else:
                return (self._curtokenpos, int(self._curtoken))
        if m.lastgroup == "keyword":
            if self._curtoken == b"true":
                return (self._curtokenpos, True)
            if self._curtoken == b"false":
                return (self._curtokenpos, False)
            return (self._curtokenpos, KWD(self._curtoken))
        if m.lastgroup == "name":
            self._curtoken = m[0][1:]
            self._curtoken = HEXDIGIT.sub(
                lambda x: bytes((int(x[1], 16),)), self._curtoken
            )
            tok = LIT(name_str(self._curtoken))
            return (self._curtokenpos, tok)
        if m.lastgroup == "startdict":
            return (self._curtokenpos, KEYWORD_DICT_BEGIN)
        if m.lastgroup == "enddict":
            return (self._curtokenpos, KEYWORD_DICT_END)
        if m.lastgroup == "startstr":
            return self._parse_endstr(self.data[self._curtokenpos + 1 : self.pos])
        if m.lastgroup == "hexstr":
            self._curtoken = SPC.sub(b"", self._curtoken[1:-1])
            if len(self._curtoken) % 2 == 1:
                self._curtoken += b"0"
            return (self._curtokenpos, unhexlify(self._curtoken))
        # Anything else is treated as a keyword (whether explicitly matched or not)
        return (self._curtokenpos, KWD(self._curtoken))

    def _parse_endstr(self, start: bytes) -> Tuple[int, Token]:
        """Parse the remainder of a string."""
        # Handle nonsense CRLF conversion in strings (PDF 1.7, p.15)
        parts = [EOLR.sub(b"\n", start)]
        paren = 1
        while True:
            m = STRLEXER.match(self.data, self.pos)
            if m is None:  # can only happen at EOS
                break
            self.pos = m.end()
            if m.lastgroup == "parenright":
                paren -= 1
                if paren == 0:
                    # By far the most common situation!
                    break
                parts.append(m[0])
                continue
            if m.lastgroup == "parenleft":
                parts.append(m[0])
                paren += 1
                continue
            if m.lastgroup == "escape":
                c = m[0][1]
                if c not in ESC_STRING:
                    # PDF 1.7 sec 7.3.4.2: If the character following
                    # the REVERSE SOLIDUS is not one of those shown in
                    # Table 3, the REVERSE SOLIDUS shall be ignored.
                    parts.append(bytes((c,)))
                else:
                    parts.append(ESC_STRING[c])
                continue
            if m.lastgroup == "octal":
                chrcode = int(m[0][1:], 8)
                if chrcode >= 256:
                    # PDF1.7 p.16: "high-order overflow shall be
                    # ignored."
                    log.warning("Invalid octal %r (%d)", m[0][1:], chrcode)
                else:
                    parts.append(bytes((chrcode,)))
                continue
            if m.lastgroup == "newline":
                # Handle nonsense CRLF conversion in strings (PDF 1.7, p.15)
                parts.append(b"\n")
                continue
            if m.lastgroup != "linebreak":
                parts.append(m[0])
        if paren != 0:
            log.warning("Unterminated string at %d", self.pos)
            raise StopIteration
        return (self._curtokenpos, b"".join(parts))


EIR: Final = re.compile(rb"\sEI\b")
EIEIR: Final = re.compile(rb"EI")
A85R: Final = re.compile(rb"\s*~\s*>\s*EI\b")
FURTHESTEIR: Final = re.compile(rb".*EI")


class ObjectParser(Iterator[Tuple[int, PDFObject]]):
    """ObjectParser is used to parse PDF object streams (and
    content streams, which have the same syntax).  Notably these
    consist of, well, a stream of objects without the surrounding
    `obj` and `endobj` tokens (which cannot occur in an object
    stream).

    They can contain indirect object references (so, must be
    initialized with a `Document` to resolve these) but for perhaps
    obvious reasons (how would you parse that) these cannot occur at
    the top level of the stream, only inside an array or dictionary.
    """

    def __init__(
        self,
        data: Union[bytes, mmap.mmap],
        doc: Union["Document", None] = None,
        pos: int = 0,
        strict: bool = False,
        streamid: Union[int, None] = None,
    ) -> None:
        self._lexer = Lexer(data, pos)
        self.stack: List[Tuple[int, PDFObject]] = []
        self.docref = None if doc is None else _ref_document(doc)
        self.strict = strict
        self.streamid = streamid

    @property
    def doc(self) -> Union["Document", None]:
        """Get associated document if it exists."""
        if self.docref is None:
            return None
        return _deref_document(self.docref)

    def newstream(
        self, data: Union[bytes, mmap.mmap], streamid: Union[int, None] = None
    ) -> None:
        """Continue parsing from a new data stream."""
        self._lexer = Lexer(data)
        self.streamid = streamid

    def reset(self) -> None:
        """Clear internal parser state."""
        del self.stack[:]

    def __next__(self) -> Any:  # should be Tuple[int, PDFObject] but mypyc
        """Get next PDF object from stream (raises StopIteration at EOF)."""
        top: Union[int, None] = None
        obj: Union[Dict[Any, Any], List[PDFObject], PDFObject] = None
        while True:
            if self.stack and top is None:
                return self.stack.pop()
            (pos, token) = next(self._lexer)
            if token is KEYWORD_ARRAY_BEGIN:
                if top is None:
                    top = pos
                self.stack.append((pos, token))
            elif token is KEYWORD_ARRAY_END:
                try:
                    pos, obj = self.pop_to(KEYWORD_ARRAY_BEGIN)
                except (TypeError, PDFSyntaxError) as e:
                    if self.strict:
                        raise e
                    log.warning("When constructing array from %r: %s", obj, e)
                if pos == top:
                    return pos, obj
                self.stack.append((pos, obj))
            elif token is KEYWORD_DICT_BEGIN:
                if top is None:
                    top = pos
                self.stack.append((pos, token))
            elif token is KEYWORD_DICT_END:
                try:
                    (pos, objs) = self.pop_to(KEYWORD_DICT_BEGIN)
                    if len(objs) % 2 != 0:
                        error_msg = (
                            "Dictionary contains odd number of objects: %r" % objs
                        )
                        raise PDFSyntaxError(error_msg)
                    obj = {
                        literal_name(k): v
                        for (k, v) in choplist(2, objs)
                        if v is not None
                    }
                except (TypeError, PDFSyntaxError) as e:
                    if self.strict:
                        raise e
                    log.warning("When constructing dict from %r: %s", self.stack, e)
                if pos == top:
                    return pos, obj
                self.stack.append((pos, obj))
            elif token is KEYWORD_PROC_BEGIN:
                if top is None:
                    top = pos
                self.stack.append((pos, token))
            elif token is KEYWORD_PROC_END:
                try:
                    pos, obj = self.pop_to(KEYWORD_PROC_BEGIN)
                except (TypeError, PDFSyntaxError) as e:
                    if self.strict:
                        raise e
                    log.warning("When constructing proc from %r: %s", obj, e)
                if pos == top:
                    return pos, obj
                self.stack.append((pos, obj))
            elif token is KEYWORD_NULL:
                self.stack.append((pos, None))
            elif token is KEYWORD_R:
                # reference to indirect object (only allowed inside another object)
                if top is None:
                    log.warning("Ignoring indirect object reference at top level")
                    self.stack.append((pos, token))
                else:
                    obj = self.get_object_reference(pos, token)
                    self.stack.append((pos, obj))
            elif token is KEYWORD_BI:
                # Inline images must occur at the top level, otherwise
                # something is wrong (probably a corrupt file)
                if top is not None:
                    raise PDFSyntaxError(
                        "Inline image not at top level of stream "
                        f"({pos} != {top}, {self.stack})"
                    )
                if (
                    self.doc is not None
                    and self.streamid is not None
                    and (inline_image_id := (self.streamid, pos))
                    in self.doc._cached_inline_images
                ):
                    end, obj = self.doc._cached_inline_images[inline_image_id]
                    self.seek(end)
                    if obj is not None:
                        return pos, obj
                else:
                    top = pos
                    self.stack.append((pos, token))
            elif token is KEYWORD_ID:
                obj = self.get_inline_image(pos, token)
                assert top is not None
                if self.doc is not None and self.streamid is not None:
                    inline_image_id = (self.streamid, top)
                    self.doc._cached_inline_images[inline_image_id] = self.tell(), obj
                if obj is not None:
                    return top, obj
            else:
                # Literally anything else, including any other keyword
                # (will be returned above if top is None, or later if
                # we are inside some object)
                self.stack.append((pos, token))

    def pop_to(self, token: PSKeyword) -> Tuple[int, List[PDFObject]]:
        """Pop everything from the stack back to token."""
        context: List[PDFObject] = []
        while self.stack:
            pos, last = self.stack.pop()
            if last is token:
                context.reverse()
                return pos, context
            context.append(last)
        raise PDFSyntaxError(f"Unmatched end token {token!r}")

    def get_object_reference(self, pos: int, token: Token) -> Union[ObjRef, None]:
        """Get an indirect object reference upon finding an "R" token."""
        _pos, genno = self.stack.pop()
        _pos, objid = self.stack.pop()
        if not isinstance(objid, int):
            if self.strict:
                raise PDFSyntaxError(
                    f"Expected object number and generation id, got {objid!r} {genno!r}"
                )
            log.warning(
                "Expected object number and generation id, got %r %r", objid, genno
            )
            return None
        if objid == 0:
            if self.strict:
                raise PDFSyntaxError(
                    "Object ID in reference at pos %d cannot be 0" % (pos,)
                )
            log.warning("Ignoring indirect object reference to 0 at %s", pos)
            return None
        return ObjRef(self.docref, objid)

    def get_inline_image(self, pos: int, token: Token) -> Union[InlineImage, None]:
        """Get an inline image upon finding an "ID" token.

        Returns a tuple of the position of the target in the data and
        the image data.  Advances the file pointer to a position after
        the "EI" token that (we hope) ends the image.

        Note: WELCOME IN THE HELL!!!
            If you're lucky enough to have PDF 2.0 documents, then you
            can skip this, you aren't actually in (the) hell.  Otherwise
            read on to know why you might be missng images or reading
            a lot of garbage in your logs:

            - The PDF 1.7 standard only specifies that image data must be
              delimited by `ID` and `EI`, and that "The bytes between
              the `ID` and `EI` operators shall be treated the same as a
              stream object’s data, even though they do not follow the
              standard stream syntax."  What does that even mean?
            - And, that must be "a single whitespace character"
              following `ID` (floating in perfume, served in a man's
              hat), except in the case of `ASCIIHexDecode` or
              `ASCII85Decode` (in which case there can just be any
              whitespace you like, or none at all).
            - It's obviously impossible to determine what a "conforming
              implementation" of this should do.

            In the easiest case, if it's `ASCIIHexDecode` data then we
            can just look for the first instance of `b"EI"`, ignoring all
            whitespace, since `b"EI"` is thankfully not a valid hex
            sequence.

            Otherwise, the stream data can, and inevitably will,
            contain the literal bytes `b"EI"`, so no, we can't just
            search for that.  In the case of `ASCII85Decode`, however,
            you can look for the `b"~>"` end-of-data sequence but note
            that sometimes it... contains whitespace!

            So, we try for `b"\\sEI\\b"`, which is not foolproof since
            you could have the pixel values `(32, 69, 73)` in your
            image followed by some other byte... so in that case,
            expect a bunch of nonsense in the logs and possible data
            loss.  Also in the rare case where `b"EI"` was preceded by
            `b"\\r\\n"`, there will be an extra `\\r` in the image
            data.  Too bad.

            And finally if that doesn't work then we will try to salvage
            something by just looking for "EI", somewhere, anywhere.  We
            take the most distant one, and if this causes you to lose
            data, well, it's definitely Adobe's fault.

            There **is** an absolutely foolproof way to parse inline
            images, but it's ridiculous so we won't do it:

            1. Find the very first instance of `b"EI"`.
            2. Extract the image itself (which could be in various formats).
            3. If it's a valid image, congratulations!  Otherwise try again.

            The moral of the story is that the author of this part of
            the PDF specification should have considered a career in
            literally anything else.

        """
        assert isinstance(token, PSKeyword) and token is KEYWORD_ID, (
            f"Not ID: {token!r}"
        )
        idpos = pos
        (pos, objs) = self.pop_to(KEYWORD_BI)
        if len(objs) % 2 != 0:
            error_msg = f"Dictionary contains odd number of objects: {objs!r}"
            if self.strict:
                raise PDFSyntaxError(error_msg)
            else:
                log.warning(error_msg)
        dic = {literal_name(k): v for (k, v) in choplist(2, objs) if v is not None}

        target_re = EIR
        whitespace_re = SPC
        # Final filter is actually the *first* in the list
        final_filter = dic.get("F", dic.get("Filter"))
        if isinstance(final_filter, list) and final_filter:
            final_filter = final_filter[0]
        if final_filter in LITERALS_ASCII85_DECODE:
            # ASCII85: look for ~>EI, ignoring all whitespace
            whitespace_re = WSR
            target_re = A85R
        elif final_filter in LITERALS_ASCIIHEX_DECODE:
            # ASCIIHex: just look for EI
            whitespace_re = WSR
            target_re = EIEIR

        # Find the start of the image data by skipping the appropriate
        # amount of whitespace.  In the case of ASCII filters, we need
        # to skip any extra whitespace before we use a possible Length
        # value (this is very dumb but the standard says...)
        pos = idpos + len(token.name)
        data = self._lexer.data
        m = whitespace_re.match(data, pos)
        if m is None:  # Note that WSR will also match nothing
            errmsg = f"ID token at {pos} not followed by whitespace"
            if self.strict:
                raise PDFSyntaxError(errmsg)
            else:
                log.warning(errmsg)
        else:
            pos = m.end(0)

        # If you have Length, you have everything
        length = dic.get("L", dic.get("Length"))
        if length is not None:
            end = pos + int_value(length)
            self.seek(end)
            (_, token) = next(self._lexer)
            if token is not KEYWORD_EI:
                errmsg = f"EI not found after Length {length!r}"
                if self.strict:
                    raise PDFSyntaxError(errmsg)
                else:
                    log.warning(errmsg)
            return InlineImage(dic, data[pos:end])

        m = target_re.search(data, pos)
        if m is not None:
            self.seek(m.end(0))
            return InlineImage(dic, data[pos : m.start(0)])
        errmsg = f"Inline image at {pos} not terminated with {target_re}"
        if self.strict:
            raise PDFSyntaxError(errmsg)
        else:
            log.warning(errmsg)

        m = FURTHESTEIR.match(data, pos)
        if m is not None:
            log.warning(
                "Inline image at %d has no whitespace before EI, "
                "expect horrible data loss!!!",
                pos,
            )
            self.seek(m.end(0))
            return InlineImage(dic, data[pos : m.end(0) - 2])
        return None

    # Delegation follows
    def seek(self, pos: int) -> None:
        """Seek to a position."""
        self._lexer.seek(pos)

    def tell(self) -> int:
        """Get the current position in the file."""
        return self._lexer.tell()

    def read(self, objlen: int) -> bytes:
        """Read data from a specified position, moving the current
        position to the end of this data."""
        return self._lexer.read(objlen)

    def nextline(self) -> Tuple[int, bytes]:
        """Read (and do not parse) next line from underlying data."""
        return self._lexer.nextline()


class IndirectObject(NamedTuple):
    objid: int
    genno: int
    obj: PDFObject


ENDSTREAMR: Final = re.compile(rb"(?:\r\n|\r|\n|)endstream")


class IndirectObjectParser(Iterator[Tuple[int, IndirectObject]]):
    """IndirectObjectParser fetches indirect objects from a data
    stream.  It holds a weak reference to the document in order to
    resolve indirect references.  If the document is deleted then this
    will obviously no longer work.

    Note that according to PDF 1.7 sec 7.5.3, "The body of a PDF file
    shall consist of a sequence of indirect objects representing the
    contents of a document."  Therefore unlike the base `ObjectParser`,
    `IndirectObjectParser` returns *only* indrect objects and not bare
    keywords, strings, numbers, etc.

    However, unlike `ObjectParser`, it will also read and return
    `ContentStream`s, as these *must* be indirect objects by definition.

    Typical usage:
      parser = IndirectObjectParser(fp, doc)
      for object in parser:
          ...

    """

    def __init__(
        self,
        data: Union[bytes, mmap.mmap],
        doc: Union["Document", None] = None,
        pos: int = 0,
        strict: bool = False,
    ) -> None:
        self._parser = ObjectParser(data, doc, pos=pos, strict=strict)
        self.buffer = data
        self.objstack: List[Tuple[int, Union[PDFObject, ContentStream]]] = []
        self.docref = None if doc is None else _ref_document(doc)
        self.strict = strict
        self.decipher = None if doc is None else doc.decipher

    @property
    def doc(self) -> Union["Document", None]:
        """Get associated document if it exists."""
        if self.docref is None:
            return None
        return _deref_document(self.docref)

    def __next__(self) -> Any:  # should be Tuple[int, IndirectObject] but mypyc
        obj: Union[PDFObject, ContentStream]
        while True:
            try:
                pos, obj = next(self._parser)
                if obj is KEYWORD_ENDOBJ:
                    return self._endobj(pos, obj)
                elif obj is KEYWORD_STREAM:
                    stream = self._stream(pos, obj)
                    self.objstack.append((pos, stream))
                elif obj is KEYWORD_ENDSTREAM:
                    if not isinstance(self.objstack[-1][1], ContentStream):
                        raise PDFSyntaxError("Got endstream without a stream")
                elif isinstance(obj, PSKeyword) and obj.name.startswith(b"endstream"):
                    # Some broken PDFs have junk after "endstream"
                    errmsg = "Expected 'endstream', got %r" % (obj,)
                    raise PDFSyntaxError(errmsg)
                elif isinstance(obj, PSKeyword) and obj.name.startswith(b"endobj"):
                    # Some broken PDFs have junk after "endobj"
                    errmsg = "Expected 'endobj', got %r" % (obj,)
                    if self.strict:
                        raise PDFSyntaxError(errmsg)
                    log.warning(errmsg)
                    return self._endobj(pos, obj)
                else:
                    self.objstack.append((pos, obj))
            except StopIteration:
                # No exceptions in strict mode!
                if self.strict:
                    raise
                # There may just have been a missing "endobj"
                if len(self.objstack) < 4:
                    raise
                # Take the object after "obj"
                for idx, (_, obj) in enumerate(self.objstack):
                    if obj is KEYWORD_OBJ:
                        break
                else:
                    # No "obj" found!
                    raise
                # No object found!
                if idx + 1 == len(self.objstack):
                    raise
                # Seek to next possible object if possible
                if idx + 2 < len(self.objstack):
                    pos, _ = self.objstack[idx + 2]
                    self._parser.seek(pos)
                del self.objstack[idx + 2 :]
                # Create an indirect object from what's left
                return self._endobj(-1, None)
            except Exception as e:
                errmsg = "Syntax error near position %d: %s" % (pos, e)
                if self.strict:
                    raise PDFSyntaxError(errmsg) from e
                else:
                    log.warning(errmsg)
                    continue

    def _endobj(self, pos: int, obj: PDFObject) -> Tuple[int, IndirectObject]:
        # Some broken PDFs omit the space after `endobj`...
        if obj is not None and obj is not KEYWORD_ENDOBJ:
            self._parser.seek(pos + len(b"endobj"))
        # objid genno "obj" (skipped) ... and the object
        (_, obj) = self.objstack.pop()
        (kpos, kwd) = self.objstack.pop()
        if kwd is not KEYWORD_OBJ:
            errmsg = "Expected 'obj' at %d, got %r" % (kpos, kwd)
            raise PDFSyntaxError(errmsg)
        (_, genno) = self.objstack.pop()
        # Update pos to be the beginning of the indirect object
        (pos, objid) = self.objstack.pop()
        try:
            objid = int_value(objid)
            genno = int_value(genno)
        except TypeError as e:
            objs = " ".join(
                [
                    repr(obj)
                    for obj in [x[1] for x in self.objstack] + [objid, genno, obj]
                ]
            )
            errmsg = (
                f"Failed to parse indirect object at {pos}: got: {objs} before 'endobj'"
            )
            raise PDFSyntaxError(errmsg) from e
        # ContentStream is *special* and needs these
        # internally for decryption.
        if isinstance(obj, ContentStream):
            obj.objid = objid
            obj.genno = genno
        # Decrypt indirect objects at top level (inside object streams
        # they are handled by ObjectStreamParser)
        if self.decipher:
            return pos, IndirectObject(
                objid,
                genno,
                decipher_all(self.decipher, objid, genno, obj),
            )
        else:
            return pos, IndirectObject(objid, genno, obj)

    def _stream(self, pos: int, obj: PDFObject) -> ContentStream:
        # PDF 1.7 sec 7.3.8.1: A stream shall consist of a
        # dictionary followed by zero or more bytes bracketed
        # between the keywords `stream` (followed by newline)
        # and `endstream`
        (_, dic) = self.objstack.pop()
        if not isinstance(dic, dict):
            # sec 7.3.8.1: the stream dictionary shall be a
            # direct object.
            raise PDFSyntaxError("Incorrect type for stream dictionary %r", dic)
        try:
            # sec 7.3.8.2: Every stream dictionary shall have
            # a Length entry that indicates how many bytes of
            # the PDF file are used for the stream’s data
            # FIXME: This call is **not** thread-safe as we currently
            # reuse the same IndirectObjectParser to resolve references
            objlen = int_value(dic["Length"])
        except KeyError:
            log.warning("/Length is undefined in stream dictionary %r", dic)
            objlen = 0
        except ValueError:
            # FIXME: This warning should be suppressed in xref
            # parsing, since we can't resolve any references yet.
            log.warning("/Length reference cannot be resolved in %r", dic)
            objlen = 0
        except TypeError:
            # FIXME: This may happen with incremental updates
            log.warning("/Length reference resolves to non-integer in %r", dic)
            objlen = 0
        # sec 7.3.8.1: The keyword `stream` that follows the stream
        # dictionary shall be followed by an end-of-line
        # marker consisting of either a CARRIAGE RETURN and a
        # LINE FEED or just a LINE FEED, and not by a CARRIAGE
        # RETURN alone.
        self._parser.seek(pos)
        _, line = self._parser.nextline()
        assert line.strip() == b"stream"
        # Because PDFs do not follow the spec, we will read
        # *at least* the specified number of bytes, which
        # could be zero (particularly if not specified!), up
        # until the "endstream" tag.  In most cases it is
        # expected that this extra data will be included in
        # the stream anyway, but for encrypted streams you
        # probably don't want that (LOL @ PDF "security")
        data = self._parser.read(objlen)
        doc = self.doc
        decipher = None if doc is None else doc.decipher
        # sec 7.3.8.1: There should be an end-of-line marker after the
        # data and before endstream; this marker shall not be included
        # in the stream length.
        #
        # TRANSLATION: We expect either one of PDF's many end-of-line
        # markers, endstream, or EOL + endstream.  If we get something
        # else, it's an error in strict mode, otherwise, we throw it
        # on the pile and keep going.
        pos = self._parser.tell()
        m = ENDSTREAMR.match(self._parser._lexer.data, pos)
        if m is not None:
            return ContentStream(dic, bytes(data), decipher)
        # We already know it's an error in strict mode, but read the
        # line anyway to show the user what's wrong
        pos, line = self._parser.nextline()
        if self.strict:
            raise PDFSyntaxError("Expected newline or 'endstream', got %r" % (line,))
        # Now glom on all the data until we see endstream
        while True:
            if b"endstream" in line:
                idx = line.index(b"endstream")
                data += line[:idx]
                self._parser.seek(pos + idx)
                break
            if b"endobj" in line:
                # Oh no! We've really gone too far now!  Stop before it gets worse
                self._parser.seek(pos)
                break
            data += line
            pos, line = self._parser.nextline()
            if line == b"":  # Means EOF
                log.warning("Incorrect length for stream, no 'endstream' found")
                break
        return ContentStream(dic, bytes(data), decipher)

    # Delegation follows
    def seek(self, pos: int) -> None:
        """Seek to a position."""
        self._parser.seek(pos)

    def tell(self) -> int:
        """Get the current position in the file."""
        return self._parser.tell()

    def reset(self) -> None:
        """Clear internal parser state."""
        self._parser.reset()


class ObjectStreamParser:
    """
    Parse indirect objects from an object stream.
    """

    def __init__(
        self,
        stream: ContentStream,
        doc: Union["Document", None] = None,
    ) -> None:
        self._parser = ObjectParser(stream.buffer, doc)
        self.buffer = stream.buffer
        self.nobj = int_value(stream["N"])
        self.first = int_value(stream["First"])
        self.offsets = []
        while True:
            try:
                _, objid = next(self._parser)
                _, pos = next(self._parser)
                objid = int_value(objid)
                pos = int_value(pos)
            except StopIteration:
                log.warning("Unexpected EOF in object stream")
                break
            self.offsets.append((objid, pos))
            if len(self.offsets) == self.nobj:
                break

    def __iter__(self) -> Iterator[Tuple[int, IndirectObject]]:
        self._parser.seek(self.first)
        for (objid, opos), (pos, obj) in zip(self.offsets, self._parser):
            if pos != self.first + opos:
                log.warning(
                    "Invalid object stream: object %d is at %d, should be at %d",
                    objid,
                    pos,
                    self.first + opos,
                )
            yield pos, IndirectObject(objid=objid, genno=0, obj=obj)


class ContentParser(Iterator[Tuple[int, PDFObject]]):
    """Parse the concatenation of multiple content streams, as
    described in the spec (PDF 1.7, p.86):

    ...the effect shall be as if all of the streams in the array were
    concatenated, in order, to form a single stream.  Conforming
    writers can create image objects and other resources as they
    occur, even though they interrupt the content stream. The division
    between streams may occur only at the boundaries between lexical
    tokens (see 7.2, "Lexical Conventions") but shall be unrelated to
    the page’s logical content or organization.
    """

    def __init__(self, streams: Iterable[PDFObject], doc: "Document") -> None:
        self.streamiter = iter(streams)
        try:
            stream = stream_value(next(self.streamiter))
            buffer = stream.buffer
            streamid = stream.objid
        except StopIteration:
            buffer = b""
            streamid = 0
        except TypeError as e:
            log.warning("Found non-stream in contents %r: %s", streams, e)
            buffer = b""
            streamid = 0
        self._parser = ObjectParser(buffer, doc, streamid=streamid)

    @property
    def streamid(self) -> Union[int, None]:
        """Indirect object ID of currently parsing stream.

        TODO: Ideally this would be returned in `__next__`, but we
        won't break the API for the moment.
        """
        return self._parser.streamid

    def __next__(self) -> Any:  # should be Tuple[int, PDFObject] but mypyc
        while True:
            try:
                return next(self._parser)
            except StopIteration:
                # Will also raise StopIteration if there are no more,
                # which is exactly what we want
                try:
                    ref = next(self.streamiter)
                    stream = stream_value(ref)
                    self._parser.newstream(stream.buffer, streamid=stream.objid)
                except TypeError:
                    log.warning("Found non-stream in contents: %r", ref)



"""
Convert an AFM file to Python font metrics.
"""

import fileinput
from pathlib import Path
from typing import Dict, TextIO

from playa.glyphlist import glyphname2unicode


def convert_font_metrics(fh: TextIO, outfh: TextIO) -> None:
    """Convert AFM files to a mapping of font metrics."""
    fonts = {}
    for line in fh:
        f = line.strip().split(" ")
        if not f:
            continue
        k = f[0]
        if k == "FontName":
            fontname = f[1].encode("latin-1")
            props = {"FontName": fontname, "Flags": 0}
            chars: Dict[str, int] = {}
            fonts[f[1]] = (props, chars)
        elif k == "C":
            cid = int(f[1])
            glyph = f[7]
            width = int(f[4])
            if glyph in glyphname2unicode:
                uni = glyphname2unicode[glyph]
            elif fontname == b"ZapfDingbats" and glyph[0] == "a":
                # FIXME: This is, really, not correct, but the whole
                # enterprise is corrupt (see
                # playa.font.Type1Font._glyph_space_width)
                uni = chr(int(glyph[1:]))
            else:
                uni = chr(cid)
            chars[uni] = width
        elif k in ("CapHeight", "XHeight", "ItalicAngle", "Ascender", "Descender"):
            k = {"Ascender": "Ascent", "Descender": "Descent"}.get(k, k)
            props[k] = float(f[1])
        elif k in ("FontName", "FamilyName", "Weight"):
            k = {"FamilyName": "FontFamily", "Weight": "FontWeight"}.get(k, k)
            props[k] = f[1].encode("latin-1")
        elif k == "IsFixedPitch":
            if f[1].lower() == "true":
                props["Flags"] = 64
        elif k == "FontBBox":
            props[k] = [float(x) for x in f[1:5]]
    print(
        '''"""Font metrics for the Adobe core 14 fonts.

Font metrics are used to compute the boundary of each character
written with a proportional font.

The following data were extracted from the AFM files:

  http://www.ctan.org/tex-archive/fonts/adobe/afm/
"""

#  BEGIN Verbatim copy of the license part
#
# Adobe Core 35 AFM Files with 314 Glyph Entries - ReadMe
#
# This file and the 35 PostScript(R) AFM files it accompanies may be
# used, copied, and distributed for any purpose and without charge,
# with or without modification, provided that all copyright notices
# are retained; that the AFM files are not distributed without this
# file; that all modifications to this file or any of the AFM files
# are prominently noted in the modified file(s); and that this
# paragraph is not modified. Adobe Systems has no responsibility or
# obligation to support the use of the AFM files.
#
#  END Verbatim copy of the license part

from typing import Dict, Tuple
from playa.pdftypes import PDFObject

FONT_METRICS: Dict[str, Tuple[Dict[str, PDFObject], Dict[str, int]]] = {
''',
        file=outfh,
    )
    for fontname, (props, chars) in fonts.items():
        print(f" {fontname!r}: {(props, chars)!r},", file=outfh)
    print(
        """}

# Aliases defined in implementation note 62 in Appecix H. related to section 5.5.1
# (Type 1 Fonts) in the PDF Reference.
FONT_METRICS["Arial"] = FONT_METRICS["Helvetica"]
FONT_METRICS["Arial,Italic"] = FONT_METRICS["Helvetica-Oblique"]
FONT_METRICS["Arial,Bold"] = FONT_METRICS["Helvetica-Bold"]
FONT_METRICS["Arial,BoldItalic"] = FONT_METRICS["Helvetica-BoldOblique"]
FONT_METRICS["CourierNew"] = FONT_METRICS["Courier"]
FONT_METRICS["CourierNew,Italic"] = FONT_METRICS["Courier-Oblique"]
FONT_METRICS["CourierNew,Bold"] = FONT_METRICS["Courier-Bold"]
FONT_METRICS["CourierNew,BoldItalic"] = FONT_METRICS["Courier-BoldOblique"]
FONT_METRICS["TimesNewRoman"] = FONT_METRICS["Times-Roman"]
FONT_METRICS["TimesNewRoman,Italic"] = FONT_METRICS["Times-Italic"]
FONT_METRICS["TimesNewRoman,Bold"] = FONT_METRICS["Times-Bold"]
FONT_METRICS["TimesNewRoman,BoldItalic"] = FONT_METRICS["Times-BoldItalic"]
""",
        file=outfh,
    )


def main():
    import argparse

    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument("afms", nargs="+", type=Path, help="AFM files to convert")
    parser.add_argument(
        "-o", "--outfile", type=argparse.FileType("w"), help="Output file"
    )
    args = parser.parse_args()
    with fileinput.input(args.afms) as fh:
        convert_font_metrics(fh, args.outfile)


if __name__ == "__main__":
    main()



"""
Get statistics on latency for all the PDFs
"""

import logging
import math
import sys
import time
from pathlib import Path

import playa
from .data import ALLPDFS, XFAILS, PASSWORDS

HERE = Path.cwd()


def main() -> None:
    open_time = []
    cat_time = []
    page_time = []
    logging.basicConfig(level=logging.ERROR)
    print("recalculating", end="", file=sys.stderr, flush=True)
    for path in ALLPDFS:
        p = Path(str(path.values[0]))
        if p.name in XFAILS:
            continue
        passwords = PASSWORDS.get(p.name, [""])
        for password in passwords:
            try:
                start = time.time()
                pdf = playa.open(p, password=password)
                open_time.append(((time.time() - start) * 1000, p.relative_to(HERE)))
                _ = pdf.catalog
                cat_time.append(((time.time() - start) * 1000, p.relative_to(HERE)))
                try:
                    _ = next(iter(pdf.pages))
                    page_time.append(
                        ((time.time() - start) * 1000, p.relative_to(HERE))
                    )
                except StopIteration:
                    pass
                print(".", end="", file=sys.stderr, flush=True)
            except playa.PDFEncryptionError:
                pass
    print(file=sys.stderr)
    report("open", open_time)
    print()
    report("catalog", cat_time)
    print()
    report("page 1", page_time)


def report(name: str, stats: list[tuple[float, Path]]) -> None:
    stats.sort(reverse=True)
    nstats = len(stats)
    print(f"{name}:")
    # We actualy do care a bit more about the mean than the median...
    mean_time = sum(t for t, p in stats) / nstats
    var_time = sum((t - mean_time) ** 2 for t, p in stats) / (nstats - 1)
    std_time = math.sqrt(var_time)
    print("  mean: %.2fms std: %.2fms" % (mean_time, std_time))
    print("  median: %.2fms %s" % stats[nstats // 2])
    print("  max: %.2fms %s" % stats[0])


if __name__ == "__main__":
    main()



"""
Benchmark ArcFour "encryption"
"""

import logging
import time
from pathlib import Path

import playa

TESTDIR = Path(__file__).parent.parent / "samples" / "encryption"

LOG = logging.getLogger(Path(__file__).stem)
PDFS = ["rc4-128.pdf", "rc4-40.pdf"]


def benchmark(path: Path):
    with playa.open(path, password="foo") as pdf:
        for page in pdf.pages:
            for stream in page.streams:
                stream.decode()


if __name__ == "__main__":
    logging.basicConfig(level=logging.ERROR)
    niter = 10
    t = 0.0
    for iter in range(niter + 1):
        for name in PDFS:
            path = TESTDIR / name
            start = time.time()
            benchmark(path)
            if iter != 0:
                t += time.time() - start
    print("arcfour took %.2f ms / iter" % (t / niter * 1000,))



"""PDF "security" handlers."""

import struct
from hashlib import md5, sha256, sha384, sha512
from typing import (
    Any,
    Callable,
    ClassVar,
    Dict,
    Final,
    FrozenSet,
    Sequence,
    Union,
)

from playa.arcfour import Arcfour
from playa.exceptions import (
    PDFEncryptionError,
    PDFPasswordIncorrect,
)
from playa.pdftypes import (
    dict_value,
    int_value,
    literal_name,
    str_value,
    uint_value,
)

default_backend: Union[Callable, None]
try:
    from cryptography.hazmat.backends import default_backend
    from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
except ImportError:
    default_backend = None

PASSWORD_PADDING = (
    b"(\xbfN^Nu\x8aAd\x00NV\xff\xfa\x01\x08..\x00\xb6\xd0h>\x80/\x0c\xa9\xfedSiz"
)


class PDFStandardSecurityHandler:
    """Basic security handler for basic encryption types."""

    supported_revisions: ClassVar[FrozenSet[int]] = frozenset((2, 3))

    def __init__(
        self,
        docid: Sequence[bytes],
        param: Dict[str, Any],
        password: str = "",
    ) -> None:
        self.docid = docid
        self.param = param
        self.password = password
        self.init()

    def init(self) -> None:
        self.init_params()
        if self.r not in self.supported_revisions:
            error_msg = "Unsupported revision: param=%r" % self.param
            raise PDFEncryptionError(error_msg)
        self.init_key()

    def init_params(self) -> None:
        self.v = int_value(self.param.get("V", 0))
        self.r = int_value(self.param["R"])
        self.p = uint_value(self.param["P"], 32)
        self.o = str_value(self.param["O"])
        self.u = str_value(self.param["U"])
        self.length = int_value(self.param.get("Length", 40))

    def init_key(self) -> None:
        self.key = self.authenticate(self.password)
        if self.key is None:
            raise PDFPasswordIncorrect

    @property
    def is_printable(self) -> bool:
        return bool(self.p & 4)

    @property
    def is_modifiable(self) -> bool:
        return bool(self.p & 8)

    @property
    def is_extractable(self) -> bool:
        return bool(self.p & 16)

    def compute_u(self, key: bytes) -> bytes:
        if self.r == 2:
            # Algorithm 3.4
            return Arcfour(key).process(PASSWORD_PADDING)  # 2
        else:
            # Algorithm 3.5
            hash = md5(PASSWORD_PADDING)  # 2
            hash.update(self.docid[0])  # 3
            result = Arcfour(key).process(hash.digest())  # 4
            for i in range(1, 20):  # 5
                k = b"".join([bytes((c ^ i,)) for c in iter(key)])
                result = Arcfour(k).process(result)
            result += result  # 6
            return result

    def compute_encryption_key(self, password: bytes) -> bytes:
        # Algorithm 3.2
        password = (password + PASSWORD_PADDING)[:32]  # 1
        hash = md5(password)  # 2
        hash.update(self.o)  # 3
        # See https://github.com/pdfminer/pdfminer.six/issues/186
        hash.update(struct.pack("<L", self.p))  # 4
        hash.update(self.docid[0])  # 5
        if self.r >= 4:
            assert isinstance(self, PDFStandardSecurityHandlerV4)
            if not self.encrypt_metadata:
                hash.update(b"\xff\xff\xff\xff")
        result = hash.digest()
        n = 5
        if self.r >= 3:
            n = self.length // 8
            for _ in range(50):
                result = md5(result[:n]).digest()
        return result[:n]

    def authenticate(self, password: str) -> Union[bytes, None]:
        password_bytes = password.encode("latin1")
        key = self.authenticate_user_password(password_bytes)
        if key is None:
            key = self.authenticate_owner_password(password_bytes)
        return key

    def authenticate_user_password(self, password: bytes) -> Union[bytes, None]:
        key = self.compute_encryption_key(password)
        if self.verify_encryption_key(key):
            return key
        else:
            return None

    def verify_encryption_key(self, key: bytes) -> bool:
        # Algorithm 3.6
        u = self.compute_u(key)
        if self.r == 2:
            return u == self.u
        return u[:16] == self.u[:16]

    def authenticate_owner_password(self, password: bytes) -> Union[bytes, None]:
        # Algorithm 3.7
        password = (password + PASSWORD_PADDING)[:32]
        hash = md5(password)
        if self.r >= 3:
            for _ in range(50):
                hash = md5(hash.digest())
        n = 5
        if self.r >= 3:
            n = self.length // 8
        key = hash.digest()[:n]
        if self.r == 2:
            user_password = Arcfour(key).process(self.o)
        else:
            user_password = self.o
            for i in range(19, -1, -1):
                k = b"".join([bytes((c ^ i,)) for c in iter(key)])
                user_password = Arcfour(k).process(user_password)
        return self.authenticate_user_password(user_password)

    def decrypt(
        self,
        objid: int,
        genno: int,
        data: bytes,
        attrs: Union[Dict[str, Any], None] = None,
    ) -> bytes:
        return self.decrypt_rc4(objid, genno, data)

    def decrypt_rc4(self, objid: int, genno: int, data: bytes) -> bytes:
        assert self.key is not None
        key = self.key + struct.pack("<L", objid)[:3] + struct.pack("<L", genno)[:2]
        hash = md5(key)
        key = hash.digest()[: min(len(key), 16)]
        return Arcfour(key).process(data)


def unpad_aes(padded: bytes) -> bytes:
    """Remove block padding as described in PDF 1.7 section 7.6.2:

    > For an original message length of M, the pad shall consist of 16 -
    (M mod 16) bytes whose value shall also be 16 - (M mod 16).
    > Note that the pad is present when M is evenly divisible by 16;
    it contains 16 bytes of 0x10.
    """
    if padded and padded[-1] <= 16:
        padding = padded[-1]
        if padding <= len(padded):
            if all(x == padding for x in padded[-padding:]):
                return padded[:-padding]
    return padded


def raise_missing_cryptography() -> None:
    if default_backend is None:
        raise PDFEncryptionError(
            "Encryption type requires the "
            "optional `cryptography` package. "
            "You may install it with `pip install playa-pdf[crypto]`."
        )


class PDFStandardSecurityHandlerV4(PDFStandardSecurityHandler):
    """Security handler for encryption type 4."""

    supported_revisions: ClassVar[FrozenSet[int]] = frozenset((4,))

    def __init__(
        self,
        docid: Sequence[bytes],
        param: Dict[str, Any],
        password: str = "",
    ) -> None:
        raise_missing_cryptography()
        super().__init__(docid, param, password)

    def init_params(self) -> None:
        super().init_params()
        self.length = 128
        self.cf = dict_value(self.param.get("CF"))
        self.stmf = literal_name(self.param["StmF"])
        self.strf = literal_name(self.param["StrF"])
        self.encrypt_metadata = bool(self.param.get("EncryptMetadata", True))
        if self.stmf != self.strf:
            error_msg = "Unsupported crypt filter: param=%r" % self.param
            raise PDFEncryptionError(error_msg)
        self.cfm = {}
        for k, v in self.cf.items():
            dictv = dict_value(v)
            f = self.get_cfm(literal_name(dictv["CFM"]))
            if f is None:
                error_msg = "Unknown crypt filter method: param=%r" % self.param
                raise PDFEncryptionError(error_msg)
            self.cfm[k] = f
        self.cfm["Identity"] = self.decrypt_identity
        if self.strf not in self.cfm:
            error_msg = "Undefined crypt filter: param=%r" % self.param
            raise PDFEncryptionError(error_msg)

    def get_cfm(self, name: str) -> Union[Callable[[int, int, bytes], bytes], None]:
        if name == "V2":
            return self.decrypt_rc4
        elif name == "AESV2":
            return self.decrypt_aes128
        else:
            return None

    def decrypt(
        self,
        objid: int,
        genno: int,
        data: bytes,
        attrs: Union[Dict[str, Any], None] = None,
        name: Union[str, None] = None,
    ) -> bytes:
        if not self.encrypt_metadata and attrs is not None:
            t = attrs.get("Type")
            if t is not None and literal_name(t) == "Metadata":
                return data
        if name is None:
            name = self.strf
        return self.cfm[name](objid, genno, data)

    def decrypt_identity(self, objid: int, genno: int, data: bytes) -> bytes:
        return data

    def decrypt_aes128(self, objid: int, genno: int, data: bytes) -> bytes:
        assert default_backend is not None
        assert self.key is not None
        key = (
            self.key
            + struct.pack("<L", objid)[:3]
            + struct.pack("<L", genno)[:2]
            + b"sAlT"
        )
        hash = md5(key)
        key = hash.digest()[: min(len(key), 16)]
        initialization_vector = data[:16]
        ciphertext = data[16:]
        cipher = Cipher(
            algorithms.AES(key),
            modes.CBC(initialization_vector),
            backend=default_backend(),
        )
        cleartext = cipher.decryptor().update(ciphertext)
        return unpad_aes(cleartext)


class PDFStandardSecurityHandlerV5(PDFStandardSecurityHandlerV4):
    """Security handler for encryption types 5 and 6."""

    supported_revisions: ClassVar[FrozenSet[int]] = frozenset((5, 6))

    def __init__(
        self,
        docid: Sequence[bytes],
        param: Dict[str, Any],
        password: str = "",
    ) -> None:
        raise_missing_cryptography()
        super().__init__(docid, param, password)

    def init_params(self) -> None:
        super().init_params()
        self.length = 256
        self.oe = str_value(self.param["OE"])
        self.ue = str_value(self.param["UE"])
        self.o_hash = self.o[:32]
        self.o_validation_salt = self.o[32:40]
        self.o_key_salt = self.o[40:]
        self.u_hash = self.u[:32]
        self.u_validation_salt = self.u[32:40]
        self.u_key_salt = self.u[40:]

    def get_cfm(self, name: str) -> Union[Callable[[int, int, bytes], bytes], None]:
        if name == "AESV3":
            return self.decrypt_aes256
        else:
            return None

    def authenticate(self, password: str) -> Union[bytes, None]:
        assert default_backend is not None
        password_b = self._normalize_password(password)
        hash = self._password_hash(password_b, self.o_validation_salt, self.u)
        if hash == self.o_hash:
            hash = self._password_hash(password_b, self.o_key_salt, self.u)
            cipher = Cipher(
                algorithms.AES(hash),
                modes.CBC(b"\0" * 16),
                backend=default_backend(),
            )
            return cipher.decryptor().update(self.oe)
        hash = self._password_hash(password_b, self.u_validation_salt)
        if hash == self.u_hash:
            hash = self._password_hash(password_b, self.u_key_salt)
            cipher = Cipher(
                algorithms.AES(hash),
                modes.CBC(b"\0" * 16),
                backend=default_backend(),
            )
            return cipher.decryptor().update(self.ue)
        return None

    def _normalize_password(self, password: str) -> bytes:
        if self.r == 6:
            # saslprep expects non-empty strings, apparently
            if not password:
                return b""
            from playa._saslprep import saslprep

            password = saslprep(password)
        return password.encode("utf-8")[:127]

    def _password_hash(
        self,
        password: bytes,
        salt: bytes,
        vector: Union[bytes, None] = None,
    ) -> bytes:
        """Compute password hash depending on revision number"""
        if self.r == 5:
            return self._r5_password(password, salt, vector)
        return self._r6_password(password, salt[0:8], vector)

    def _r5_password(
        self,
        password: bytes,
        salt: bytes,
        vector: Union[bytes, None] = None,
    ) -> bytes:
        """Compute the password for revision 5"""
        hash = sha256(password)
        hash.update(salt)
        if vector is not None:
            hash.update(vector)
        return hash.digest()

    def _r6_password(
        self,
        password: bytes,
        salt: bytes,
        vector: Union[bytes, None] = None,
    ) -> bytes:
        """Compute the password for revision 6"""
        initial_hash = sha256(password)
        initial_hash.update(salt)
        if vector is not None:
            initial_hash.update(vector)
        k = initial_hash.digest()
        hashes = (sha256, sha384, sha512)
        round_no = last_byte_val = 0
        while round_no < 64 or last_byte_val > round_no - 32:
            k1 = (password + k + (vector or b"")) * 64
            e = self._aes_cbc_encrypt(key=k[:16], iv=k[16:32], data=k1)
            # compute the first 16 bytes of e,
            # interpreted as an unsigned integer mod 3
            next_hash = hashes[self._bytes_mod_3(e[:16])]
            k = next_hash(e).digest()
            last_byte_val = e[len(e) - 1]
            round_no += 1
        return k[:32]

    @staticmethod
    def _bytes_mod_3(input_bytes: bytes) -> int:
        # 256 is 1 mod 3, so we can just sum 'em
        return sum(b % 3 for b in input_bytes) % 3

    def _aes_cbc_encrypt(self, key: bytes, iv: bytes, data: bytes) -> bytes:
        cipher = Cipher(algorithms.AES(key), modes.CBC(iv))
        encryptor = cipher.encryptor()
        return encryptor.update(data) + encryptor.finalize()

    def decrypt_aes256(self, objid: int, genno: int, data: bytes) -> bytes:
        assert default_backend is not None
        initialization_vector = data[:16]
        ciphertext = data[16:]
        assert self.key is not None
        cipher = Cipher(
            algorithms.AES(self.key),
            modes.CBC(initialization_vector),
            backend=default_backend(),
        )
        cleartext = cipher.decryptor().update(ciphertext)
        return unpad_aes(cleartext)


SECURITY_HANDLERS: Final = {
    1: PDFStandardSecurityHandler,
    2: PDFStandardSecurityHandler,
    4: PDFStandardSecurityHandlerV4,
    5: PDFStandardSecurityHandlerV5,
}



"""
Test basic opening and navigation of PDF documents.
"""

import pytest

import playa

from .data import CONTRIB, TESTDIR


def test_weakrefs() -> None:
    """Verify that PDFDocument really gets deleted even if we have
    PDFObjRefs hanging around."""
    with playa.open(TESTDIR / "simple5.pdf") as doc:
        ref = doc.catalog["Pages"]
    del doc
    with pytest.raises(RuntimeError):
        _ = ref.resolve()


@pytest.mark.skipif(not CONTRIB.exists(), reason="contrib samples not present")
def test_spaces() -> None:
    """Test different coordinate spaces."""
    with playa.open(CONTRIB / "issue-1181.pdf", space="page") as doc:
        page = doc.pages[0]
        page_box = next(iter(page)).bbox
        assert page_box is not None
    with playa.open(CONTRIB / "issue-1181.pdf", space="default") as doc:
        page = doc.pages[0]
        user_box = next(iter(page)).bbox
        assert user_box is not None
    assert page_box[1] == pytest.approx(user_box[1] - page.mediabox[1])
    with playa.open(CONTRIB / "issue-1181.pdf", space="screen") as doc:
        page = doc.pages[0]
        screen_box = next(iter(page)).bbox
    # BBoxes are normalied, so top is 1 for screen and 3 for page
    assert screen_box is not None
    assert screen_box[3] == pytest.approx(page.height - page_box[1])
    assert screen_box[3] == pytest.approx(page.height - page_box[1])


def test_tiff_predictor() -> None:
    with playa.open(TESTDIR / "test_pdf_with_tiff_predictor.pdf") as doc:
        image = next(doc.pages[0].images)
        # Decoded TIFF: 600 x 600 + a header
        assert len(image.stream.buffer) == 360600



"""
Test the actual parser.
"""

import logging

import pytest
from playa.document import Document
from playa.exceptions import PDFSyntaxError
from playa.parser import InlineImage, ObjectParser
from playa.pdftypes import KWD, LIT, ContentStream, ObjRef

logger = logging.getLogger(__name__)

STREAMDATA = b"""
/Hello
<< /Type/Catalog/Outlines 2 0 R /Pages 3 0 R >>
[ 1 0 R ]
(foo (bar) baz...)
null null null
4 0 R
"""


def test_objects():
    """Test the basic object stream parser."""
    parser = ObjectParser(STREAMDATA)
    objects = list(parser)
    assert objects == [
        (1, LIT("Hello")),
        (
            8,
            {
                "Type": LIT("Catalog"),
                "Outlines": ObjRef(None, 2),
                "Pages": ObjRef(None, 3),
            },
        ),
        (56, [ObjRef(None, 1)]),
        (66, b"foo (bar) baz..."),
        (85, None),
        (90, None),
        (95, None),
        # Note unparsed indirect object reference
        (100, 4),
        (102, 0),
        (104, KWD(b"R")),
    ]


INLINEDATA1 = b"""
BI /L 42 ID
012345678901234567890123456789012345678901
EI
BI /Length 30 /Filter /A85 ID


<^BVT:K:=9<E)pd;BS_1:/aSV;ag~>



EI
BI
/Foo (bar)
ID
VARIOUS UTTER NONSENSE
EI
BI
/F /AHx
ID\r4f 4d 47575446\r\n\r\nEI
BI /F /AHx ID 4f4d47575446EI
BI ID(OMG)(WTF)
EI
BI
/F /A85
ID
<^BVT:K:=9<E)pd;BS_1:/aSV;ag~>
EI
BI
/F /A85
ID
<^BVT:K:=9<E)pd;BS_1:/aSV;ag~
>
EI
BI /F /A85 ID<^BVT:K:=9<E)pd;BS_1:/aSV;ag~>EI
BI
/OMG (WTF)
ID
BLAHEIBLAHBLAH\rEI
BI ID
OLD MACDONALD\rEIEIO
EI
BI ID OLDMACDONALDEIEIO EI
BI ID
OLDMACDONALDEIEIOEI
(hello world)
"""


def test_inline_images():
    parser = ObjectParser(INLINEDATA1)
    pos, img = next(parser)
    assert isinstance(img, InlineImage)
    assert img.buffer == b"012345678901234567890123456789012345678901"
    pos, img = next(parser)
    assert isinstance(img, InlineImage)
    assert img.buffer == b"VARIOUS UTTER NONSENSE"
    pos, img = next(parser)
    assert isinstance(img, InlineImage)
    assert img.attrs["Foo"] == b"bar"
    assert img.rawdata == b"VARIOUS UTTER NONSENSE"
    pos, img = next(parser)
    assert isinstance(img, InlineImage)
    assert img.buffer == b"OMGWTF"
    pos, img = next(parser)
    assert isinstance(img, InlineImage)
    assert img.buffer == b"OMGWTF"
    pos, img = next(parser)
    assert isinstance(img, InlineImage)
    assert img.buffer == b"(OMG)(WTF)"
    pos, img = next(parser)
    assert isinstance(img, InlineImage)
    assert img.buffer == b"VARIOUS UTTER NONSENSE"
    pos, img = next(parser)
    assert isinstance(img, InlineImage)
    assert img.buffer == b"VARIOUS UTTER NONSENSE"
    pos, img = next(parser)
    assert isinstance(img, InlineImage)
    assert img.buffer == b"VARIOUS UTTER NONSENSE"
    pos, img = next(parser)
    assert isinstance(img, InlineImage)
    assert img.buffer == b"BLAHEIBLAHBLAH"
    pos, img = next(parser)
    assert isinstance(img, InlineImage)
    assert img.buffer == b"OLD MACDONALD\rEIEIO"
    pos, img = next(parser)
    assert isinstance(img, InlineImage)
    assert img.buffer == b"OLDMACDONALDEIEIO"
    pos, img = next(parser)
    assert isinstance(img, InlineImage)
    assert img.buffer == b"OLDMACDONALDEIEIO"


def test_cached_inline_images():
    doc = Document(b"trailer << >>")
    first = list(ObjectParser(INLINEDATA1, doc, streamid=0))
    second = list(ObjectParser(INLINEDATA1, doc, streamid=0))
    assert all(
        a is b for (_, a), (_, b) in zip(first, second) if isinstance(a, ContentStream)
    )
    third = list(ObjectParser(INLINEDATA1, doc, streamid=1))
    assert any(
        a is not b
        for (_, a), (_, b) in zip(first, third)
        if isinstance(a, ContentStream)
    )


def test_strict_errors() -> None:
    """Verify that the strict parser is strict."""
    with pytest.raises(PDFSyntaxError):
        list(ObjectParser(b"1 2 3]", strict=True))
    with pytest.raises(PDFSyntaxError):
        list(ObjectParser(b"<< /Foo (bar) /Baz >>", strict=True))
    with pytest.raises(TypeError):
        list(ObjectParser(b"<< 32 /WTF >>", strict=True))
    with pytest.raises(PDFSyntaxError):
        list(ObjectParser(b"1 2 3}", strict=True))
    with pytest.raises(PDFSyntaxError):
        list(ObjectParser(b"[ 1 2 BI", strict=True))
    with pytest.raises(PDFSyntaxError):
        list(ObjectParser(b"[ 0 0 R ]", strict=True))
    with pytest.raises(PDFSyntaxError):
        list(ObjectParser(b"[ 2 R ]", strict=True))
    with pytest.raises(PDFSyntaxError):
        list(
            ObjectParser(
                b"""BI /L 42 /WTF ID
012345678901234567890123456789012345678901
EI
        """,
                strict=True,
            )
        )
    with pytest.raises(PDFSyntaxError):
        list(
            ObjectParser(
                b"""BI /L 20 ID
012345678901234567890123456789012345678901
EI
        """,
                strict=True,
            )
        )
    with pytest.raises(PDFSyntaxError):
        list(
            ObjectParser(
                b"""BI ID
012345678901234567890123456789012345678901
        """,
                strict=True,
            )
        )
    with pytest.raises(PDFSyntaxError):
        # ID must be followed by whitespace when ASCII filters not used
        list(
            ObjectParser(
                b"""BI ID/12345678901234567890123456789012345678901 EI""", strict=True
            )
        )


def test_warn_errors() -> None:
    """Invoke various warnings."""
    list(ObjectParser(b"<< /Foo (bar) /Baz >>"))
    list(ObjectParser(b"<< 32 /WTF >>"))
    list(ObjectParser(b"[ 2 R ]"))
    list(
        ObjectParser(
            b"""BI /L 42 /WTF ID
012345678901234567890123456789012345678901
EI
    """
        )
    )
    list(
        ObjectParser(
            b"""BI /L 20 ID
012345678901234567890123456789012345678901
EI
    """
        )
    )
    list(
        ObjectParser(
            b"""BI ID
012345678901234567890123456789012345678901
    """
        )
    )



"""
Convert glyph list into Python code.
"""

from typing import TextIO


def convert_glyphlist(fileinput: TextIO) -> None:
    """Convert a glyph list into a python representation."""
    state = 0
    for line in fileinput:
        line = line.strip()
        if not line or line.startswith("#"):
            if state == 1:
                state = 2
                print("}\n")
            print(line)
            continue
        if state == 0:
            print("\nglyphname2unicode = {")
            state = 1
        (name, x) = line.split(";")
        codes = x.split(" ")
        print(
            " {!r}: u'{}',".format(name, "".join("\\u%s" % code for code in codes)),
        )


def main():
    import argparse

    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        "glyphlist", type=argparse.FileType("r"), help="Adobe glyph list to convert"
    )
    args = parser.parse_args()
    convert_glyphlist(args.glyphlist)


if __name__ == "__main__":
    main()



"""
Benchmark marked content section iteration and extraction.
"""

import logging
import time
from pathlib import Path

import playa

SAMPLES = Path(__file__).parent.parent / "samples"
CONTRIB = Path(__file__).parent.parent / "samples" / "contrib"

LOG = logging.getLogger("benchmark-text")
# Use a standard benchmark set to make version comparisons possible
PDFS = [
    "2023-04-06-ODJ et Résolutions-séance xtra 6 avril 2023.pdf",
    "2023-06-20-PV.pdf",
    "PSC_Station.pdf",
    "Rgl-1314-2021-DM-Derogations-mineures.pdf",
]


def benchmark_cli(path: Path) -> None:
    with playa.open(path) as doc:
        for page in doc.pages:
            for _mcid, element in enumerate(page.structure):
                if element is None:
                    continue
                _ = element.bbox


if __name__ == "__main__":
    logging.basicConfig(level=logging.ERROR)
    niter = 5
    cli_time = 0.0
    for iter in range(niter + 1):
        for name in PDFS:
            path = SAMPLES / name
            if not path.exists():
                path = CONTRIB / name
            start = time.time()
            benchmark_cli(path)
            if iter != 0:
                cli_time += time.time() - start
    print("MCS access took %d ms / iter" % (cli_time / niter * 1000,))



"""
Test the CLI
"""

from contextlib import redirect_stdout
from pathlib import Path

import pytest
import tempfile

from playa import PDFPasswordIncorrect
from playa.cli import main
from playa.exceptions import PDFEncryptionError
from tests.data import ALLPDFS, PASSWORDS, XFAILS, TESTDIR


@pytest.mark.parametrize("path", ALLPDFS, ids=str)
def test_cli_metadata(path: Path):
    if path.name in XFAILS:
        pytest.xfail("Expected failure: %s" % path.name)
    passwords = PASSWORDS.get(path.name, [""])
    for password in passwords:
        try:
            # FIXME: Verify that output is valid JSON
            main(["--password", password, "--non-interactive", str(path)])
        except PDFPasswordIncorrect:
            pass
        except PDFEncryptionError:
            pytest.skip("cryptography package not installed")


@pytest.mark.parametrize("path", ALLPDFS, ids=str)
def test_cli_catalog(path: Path):
    if path.name in XFAILS:
        pytest.xfail("Expected failure: %s" % path.name)
    passwords = PASSWORDS.get(path.name, [""])
    for password in passwords:
        try:
            # FIXME: Verify that output is valid JSON
            main(["--password", password, "--catalog", "--non-interactive", str(path)])
        except PDFPasswordIncorrect:
            pass
        except PDFEncryptionError:
            pytest.skip("cryptography package not installed")


@pytest.mark.parametrize("path", ALLPDFS, ids=str)
def test_cli_outline(path: Path):
    if path.name in XFAILS:
        pytest.xfail("Expected failure: %s" % path.name)
    passwords = PASSWORDS.get(path.name, [""])
    for password in passwords:
        try:
            # FIXME: Verify that output is valid JSON
            main(["--password", password, "--non-interactive", "--outline", str(path)])
        except PDFPasswordIncorrect:
            pass
        except PDFEncryptionError:
            pytest.skip("cryptography package not installed")


@pytest.mark.parametrize("path", ALLPDFS, ids=str)
def test_cli_structure(path: Path):
    if path.name in XFAILS:
        pytest.xfail("Expected failure: %s" % path.name)
    passwords = PASSWORDS.get(path.name, [""])
    for password in passwords:
        try:
            # FIXME: Verify that output is valid JSON
            main(
                ["--password", password, "--non-interactive", "--structure", str(path)]
            )
        except PDFPasswordIncorrect:
            pass
        except PDFEncryptionError:
            pytest.skip("cryptography package not installed")


@pytest.mark.parametrize("path", ALLPDFS, ids=str)
def test_cli_text(path: Path):
    if path.name in XFAILS:
        pytest.xfail("Expected failure: %s" % path.name)
    passwords = PASSWORDS.get(path.name, [""])
    for password in passwords:
        try:
            main(["--password", password, "--non-interactive", "--text", str(path)])
        except PDFPasswordIncorrect:
            pass
        except PDFEncryptionError:
            pytest.skip("cryptography package not installed")


@pytest.mark.parametrize("path", ALLPDFS, ids=str)
def test_cli_content_objects(path: Path):
    if path.name in XFAILS:
        pytest.xfail("Expected failure: %s" % path.name)
    passwords = PASSWORDS.get(path.name, [""])
    for password in passwords:
        try:
            # Avoid OOM errors from capturing too much output
            with redirect_stdout(None):
                main(
                    [
                        "--password",
                        password,
                        "--non-interactive",
                        "--content-objects",
                        str(path),
                    ]
                )
        except PDFPasswordIncorrect:
            pass
        except PDFEncryptionError:
            pytest.skip("cryptography package not installed")


@pytest.mark.parametrize("path", ALLPDFS, ids=str)
def test_cli_images(path: Path):
    if path.name in XFAILS:
        pytest.xfail("Expected failure: %s" % path.name)
    passwords = PASSWORDS.get(path.name, [""])
    for password in passwords:
        try:
            # FIXME: Verify that output is valid JSON
            with tempfile.TemporaryDirectory() as tempdir:
                main(
                    [
                        "--password",
                        password,
                        "--non-interactive",
                        "--images",
                        tempdir,
                        str(path),
                    ]
                )
        except PDFPasswordIncorrect:
            pass
        except PDFEncryptionError:
            pytest.skip("cryptography package not installed")


@pytest.mark.parametrize("path", ALLPDFS, ids=str)
def test_cli_fonts(path: Path):
    if path.name in XFAILS:
        pytest.xfail("Expected failure: %s" % path.name)
    passwords = PASSWORDS.get(path.name, [""])
    for password in passwords:
        try:
            # FIXME: Verify that output is valid JSON
            with tempfile.TemporaryDirectory() as tempdir:
                main(
                    [
                        "--password",
                        password,
                        "--non-interactive",
                        "--fonts",
                        tempdir,
                        str(path),
                    ]
                )
        except PDFPasswordIncorrect:
            pass
        except PDFEncryptionError:
            pytest.skip("cryptography package not installed")


def test_extract_stream():
    """Verify that stream extractor works right."""
    main(["-t", "2", str(TESTDIR / "pdf_structure.pdf")])


def test_extract_catalog():
    """Verify that catalog extractor works right."""
    main(["--catalog", str(TESTDIR / "pdf_structure.pdf")])


def test_page_specs():
    """Verify page specifications."""
    testpdf = str(TESTDIR / "font-size-test.pdf")
    main(["--pages", "all", "--text", testpdf])
    main(["--pages", "all,1,3", "--text", testpdf])
    main(["--pages", "1-2", "--text", testpdf])
    main(["--pages", "1-2,4,5", "--text", testpdf])
    main(["--pages", "10-4", "--text", testpdf])
    main(["--pages", "666-668", "--text", testpdf])
    main(["--pages", "668-666", "--text", testpdf])


def test_no_args():
    with pytest.raises(SystemExit):
        main([])


def test_bad_page_spec():
    testpdf = str(TESTDIR / "font-size-test.pdf")
    assert main(["--pages", "10-4,goodbuddy", "--text", testpdf]) != 0


def test_text_objects():
    testpdf = str(TESTDIR / "font-size-test.pdf")
    main(["--pages", "1,3", "--text-objects", testpdf])


def test_content_objects():
    testpdf = str(TESTDIR / "font-size-test.pdf")
    main(["--pages", "1,3", "--content-objects", testpdf])
    main(["--pages", "1,3", "--content-objects", "--explode-text", testpdf])


def test_content_streams():
    testpdf = str(TESTDIR / "font-size-test.pdf")
    main(["--pages", "1", "--content-streams", testpdf])



"""
Test various font-related things
"""

import json
from typing import List

import playa
import pytest
from playa.content import PathObject
from playa.font import Type3Font
from playa.pdftypes import Rect, dict_value
from playa.utils import get_bound_rects

from .data import CONTRIB, TESTDIR


def test_implicit_encoding_type1() -> None:
    """Test implicit encodings for Type1 fonts."""
    with playa.open(TESTDIR / "simple5.pdf") as doc:
        page = doc.pages[0]
        fonts = page.resources.get("Font")
        assert fonts is not None
        assert isinstance(fonts, dict)
        for name, desc in fonts.items():
            font = doc.get_font(desc.objid, desc.resolve())
            assert font is not None
            if 147 in font.encoding:
                assert font.encoding[147] == "quotedblleft"


def test_custom_encoding_core() -> None:
    """Test custom encodings for core fonts."""
    with playa.open(TESTDIR / "core_font_encodings.pdf") as doc:
        page = doc.pages[0]
        # Did we get the encoding right? (easy)
        assert (
            page.extract_text_untagged()
            == """\
Ç’est ça mon Bob
Un peu plus à droite"""
        )
        # Did we get the *glyphs* right? (harder)
        boxes = list(t.bbox for t in page.texts)
        assert boxes[0] == pytest.approx((100.0, 74.768, 289.408, 96.968))
        assert boxes[1] == pytest.approx((150.0, 110.768, 364.776, 132.968))


@pytest.mark.skipif(not CONTRIB.exists(), reason="contrib samples not present")
def test_implicit_encoding_cff() -> None:
    with playa.open(CONTRIB / "implicit_cff_encoding.pdf") as doc:
        page = doc.pages[0]
        fonts = page.resources.get("Font")
        assert fonts is not None
        assert isinstance(fonts, dict)
        for name, desc in fonts.items():
            font = doc.get_font(desc.objid, desc.resolve())
            assert font.encoding
        # Verify fallback to StandardEncoding
        t = page.extract_text()
        assert t.strip() == "Part I\nClick here to access Part II \non hp.com."


@pytest.mark.skipif(not CONTRIB.exists(), reason="contrib samples not present")
def test_implicit_encoding_cff_issue91() -> None:
    """Ensure that we can properly parse some CFF programs."""
    with playa.open(CONTRIB / "issue-91.pdf") as doc:
        page = doc.pages[0]
        fonts = page.resources.get("Font")
        assert fonts is not None
        assert isinstance(fonts, dict)
        for name, desc in fonts.items():
            font = doc.get_font(desc.objid, desc.resolve())
            # Font should have an encoding
            assert font.encoding
            # It should *not* be the standard encoding
            assert 90 not in font.encoding


def test_type3_font_boxes() -> None:
    """Ensure that we get bounding boxes right for Type3 fonts with
    mildly exotic FontMatrix"""
    with playa.open(TESTDIR / "type3_fonts.pdf") as doc:
        font = doc.get_font(5, dict_value(doc[5]))
        assert font.basefont == "BAAAAA+Open-Sans-Light"
        assert font.fontname == "BAAAAA+Open-Sans-Light"
        # This font's BBox is really something
        assert font.bbox == (-164, 493, 1966, -1569)
        assert isinstance(font, Type3Font)
        assert font.matrix == (0.0004882813, 0, 0, -0.0004882813, 0, 0)
        page = doc.pages[0]
        textor = page.texts
        line1 = next(textor).bbox
        assert line1 == pytest.approx((25.0, 14.274413, 246.586937, 28.370118))
        boxes: List[Rect] = []
        for text in textor:
            bbox = text.bbox
            assert bbox is not None
            # They should be mostly adjacent and aligned
            if boxes:
                assert bbox[0] == pytest.approx(boxes[-1][2])
                assert bbox[1] == pytest.approx(boxes[-1][1])
                assert bbox[3] == pytest.approx(boxes[-1][3])
            boxes.append(bbox)
        line2 = get_bound_rects(boxes)
        assert line2 == pytest.approx(
            (25.0, 39.274413, 246.58691507160006, 53.3701175326)
        )


def test_exotic_type3_font_boxes() -> None:
    """Ensure that we get bounding boxes right for Type3 fonts with
    seriously exotic FontMatrix"""
    with playa.open(TESTDIR / "rotated_type3_fonts.pdf") as doc:
        page = doc.pages[0]
        f30 = page.fonts["F5R30"]
        assert f30.matrix == (
            0.000422864,
            -0.000244141,
            -0.000244141,
            -0.000422864,
            0,
            0,
        )
        f30r = page.fonts["F5RM30"]
        assert f30r.matrix == (
            0.000422864,
            0.000244141,
            0.000244141,
            -0.000422864,
            0,
            0,
        )
        # Ensure char bboxes take rotation into account (the fonts are
        # othewise the same)
        assert f30.char_bbox(0) != f30r.char_bbox(0)
        # Ensure TextObject bboxes take rotation into account
        boxes = list(t.bbox for t in page.texts if t.bbox is not None)
        # Rotating backwards moves the left side backwards
        assert boxes[1][0] < boxes[0][0]
        # Rotating forwards moves the right side forwards
        assert boxes[0][2] > boxes[1][2]
        # Height remains (at least approximately) the same
        assert (boxes[0][3] - boxes[0][1]) == pytest.approx(boxes[1][3] - boxes[1][1])


@pytest.mark.parametrize(
    "name",
    ["vertical_writing", "vertical_writing_offset", "simple3", "character_spacing"],
)
def test_glyph_positioning(name: str) -> None:
    """Verify that various more or less exotic aspects of glyph
    positioning are handled correctly."""
    with open(TESTDIR / f"{name}_texts.json", encoding="utf-8") as infh:
        texts = json.load(infh)
    with open(TESTDIR / f"{name}_glyphs.json", encoding="utf-8") as infh:
        glyphs = json.load(infh)
    with playa.open(TESTDIR / f"{name}.pdf", space="default") as doc:
        page = doc.pages[0]
        for text, expected in zip(page.texts, texts):
            assert text.chars == expected["chars"]
            assert text.bbox == tuple(expected["bbox"])
        for glyph, expected in zip(page.glyphs, glyphs):
            assert glyph.text == expected["text"]
            assert glyph.bbox == tuple(expected["bbox"])


@pytest.mark.skipif(not CONTRIB.exists(), reason="contrib samples not present")
def test_fallback_type3_cid2unicode() -> None:
    """Verify that we fall back to standard encoding when a Type3 font
    has nonsense glyph names."""
    with playa.open(CONTRIB / "anonymous_type3_fonts.pdf") as doc:
        wtf = doc.pages[2].fonts["T3_0"]
        # Additionally verify that we correctly use Name as FontName
        # and (if not obviously subset) BaseFont for Type3 fonts
        assert wtf.basefont == "F47"
        assert wtf.fontname == "F47"
        assert list(wtf.decode(b"scienti\xaec")) == [
            (115, "s"),
            (99, "c"),
            (105, "i"),
            (101, "e"),
            (110, "n"),
            (116, "t"),
            (105, "i"),
            (174, "ﬁ"),
            (99, "c"),
        ]


def test_type3_charprocs() -> None:
    """Test iteration over Type3 font programs."""
    with playa.open(TESTDIR / "type3_fonts.pdf") as pdf:
        f5 = pdf.pages[0].fonts["F5"]
        assert isinstance(f5, Type3Font)
        assert f5.resources is None
        assert "g26" in f5.charprocs
        c = next(iter(pdf.pages[0].glyphs))
        assert c.cid == 38
        assert c.text == "C"
        p = next(iter(c))
        assert isinstance(p, PathObject)
        assert p.ctm == pytest.approx(
            (0.0068359382, 0.0, 0.0, 0.0068359382, 25.0, 25.0)
        )
        print(list(p.segments))


def test_bogus_metrics() -> None:
    """Verify that we fix bogus font descriptors with ascent = descent = 0."""
    with playa.open(TESTDIR / "pdf_structure.pdf") as pdf:
        for t in pdf.pages[0].texts:
            assert t.bbox is not None
            x0, y0, x1, y1 = t.bbox
            assert y1 > y0



"""
Some PLAYA-specific exceptions to cover PDF-specific problems
"""


class PDFException(Exception):
    pass


class PDFSyntaxError(PDFException):
    pass


class PDFInterpreterError(PDFException):
    pass


class PDFEncryptionError(PDFException):
    pass


class PDFPasswordIncorrect(PDFEncryptionError):
    pass


class PDFTextExtractionNotAllowed(PDFEncryptionError):
    pass


class PDFFontError(PDFException):
    pass


class PDFUnicodeNotDefined(PDFFontError):
    pass



"""PDF cross-reference tables / streams."""

import itertools
import logging
import re
from typing import (
    Dict,
    Final,
    Iterator,
    List,
    Mapping,
    NamedTuple,
    Tuple,
    Union,
    TYPE_CHECKING,
)

from playa.exceptions import (
    PDFSyntaxError,
)
from playa.parser import (
    KEYWORD_TRAILER,
    LIT,
    IndirectObjectParser,
    ObjectParser,
)
from playa.pdftypes import (
    ContentStream,
    ObjRef,
    PDFObject,
    dict_value,
    int_value,
    list_value,
    stream_value,
)
from playa.utils import (
    choplist,
    nunpack,
)
from playa.worker import _ref_document

if TYPE_CHECKING:
    from playa.document import Document

log: Final = logging.getLogger(__name__)
LITERAL_OBJSTM: Final = LIT("ObjStm")
LITERAL_XREF: Final = LIT("XRef")
# Specific regex optimized only for finding objects (SFOOFFO)
FIND_INDOBJR: Final = re.compile(rb"(?<!\d)\d{1,10}\s+\d{1,10}\s+obj")
INDOBJR: Final = re.compile(rb"\s*\d{1,10}\s+\d{1,10}\s+obj")
XREFR: Final = re.compile(rb"\s*xref\s*(\d+)\s*(\d+)\s*")
ORDN: Final = ord(b"n")
ORDF: Final = ord(b"f")


def _update_refs(trailer: Dict[str, PDFObject], doc: "Document") -> None:
    docref = _ref_document(doc)
    for val in trailer.values():
        if isinstance(val, ObjRef):
            val.doc = docref


class XRefPos(NamedTuple):
    streamid: Union[int, None]
    pos: int
    genno: int


class XRef(Mapping[int, XRefPos]):
    """
    XRef table interface (expected to be read-only)
    """

    trailer: Dict[str, PDFObject]


class XRefTable(XRef):
    """Simplest (PDF 1.0) implementation of cross-reference table, in
    plain text at the end of the file.
    """

    def __init__(
        self, doc: Union["Document", None] = None, pos: int = 0, offset: int = 0
    ) -> None:
        self.subsections: List[XRefTableSubsection] = []
        self.trailer: Dict[str, PDFObject] = {}
        if doc is not None:
            self._load(ObjectParser(doc.buffer, doc, pos), offset)

    def _load(self, parser: ObjectParser, offset: int) -> None:
        while True:
            pos, start = next(parser)
            # This means that xref table parsing can only end in three
            # ways: "trailer" (success), EOF (failure) or something
            # other than two numbers (failure).  Hope that's okay.
            if start is KEYWORD_TRAILER:
                parser.seek(pos)
                break
            pos, nobjs = next(parser)
            if not (isinstance(start, int) and isinstance(nobjs, int)):
                raise PDFSyntaxError(
                    f"Expected object ID and count, got {start!r} {nobjs!r}"
                )
            # Cue up the table data by skipping any whitespace
            pos, ws = parser.nextline()
            if ws.strip():  # Not whitespace, it's an error
                raise PDFSyntaxError(f"Unexpected data in xref table: {ws!r}")
            table_data = parser.read(20 * nobjs)
            if len(table_data) != 20 * nobjs:
                raise PDFSyntaxError(f"EOF in reading xref table data at {pos}")
            subsection = XRefTableSubsection(table_data, start, nobjs, offset)
            if log.isEnabledFor(logging.DEBUG):
                log.debug(subsection)
                for objid in subsection:
                    try:
                        ref = subsection[objid]
                    except Exception as e:
                        raise PDFSyntaxError from e
                    log.debug("object %d %d at pos %d", objid, ref.genno, ref.pos)
            self.subsections.append(subsection)
        self._load_trailer(parser)

    def _load_trailer(self, parser: ObjectParser) -> None:
        (_, kwd) = next(parser)
        # This can actually never happen, because if an xref table
        # doesn't end with "trailer" then some other error happens
        if kwd is not KEYWORD_TRAILER:
            raise PDFSyntaxError(
                "Expected %r, got %r"
                % (
                    KEYWORD_TRAILER,
                    kwd,
                )
            )
        (_, dic) = next(parser)
        self.trailer.update(dict_value(dic))

    def __repr__(self) -> str:
        return "<XRefTable: subsections=%r>" % (self.subsections)

    def __len__(self) -> int:
        return sum(len(x) for x in self.subsections)

    def __iter__(self) -> Iterator[int]:
        return itertools.chain.from_iterable(self.subsections)

    def __getitem__(self, objid: int) -> XRefPos:
        for s in self.subsections:
            if objid in s.range:
                return s[objid]
        raise KeyError


class XRefTableSubsection:
    """A contiguous chunk of object numbers from the cross-reference table."""

    def __init__(
        self,
        data: bytes = b"",
        start: int = 0,
        nobjs: int = 0,
        offset: int = 0,
    ):
        self.data = data
        self.start = start
        self.nobjs = nobjs
        self.offset = offset

    def __repr__(self) -> str:
        return "<XRefTableSubsection: start=%d, nobjs=%d>" % (self.start, self.nobjs)

    @property
    def range(self) -> range:
        return range(self.start, self.start + self.nobjs)

    def __len__(self):
        return self.nobjs

    def __iter__(self):
        for objid in self.range:
            if objid in self:
                yield objid

    def __contains__(self, objid):
        return self._get_row(objid)[17] != ORDF

    def _get_row(self, objid: int) -> bytes:
        table_offset = 20 * self.range.index(objid)
        return self.data[table_offset : table_offset + 20]

    def __getitem__(self, objid: int) -> XRefPos:
        row = self._get_row(objid)
        pos = int(row[0:10])
        genno = int(row[11:16])
        use = row[17]
        if use == ORDN:
            return XRefPos(None, pos + self.offset, genno)
        if use == ORDF:
            # NOTE: PDF standard says this is `null` but we will raise
            # a KeyError in keeping with Python protocols
            raise KeyError(f"Object ID {objid} does not exist")
        raise PDFSyntaxError(f"Invalid xref table entry: {row!r}")


class XRefFallback(XRef):
    """In the case where a file is non-conforming and has no
    `startxref` marker at its end, we will reconstruct a
    cross-reference table by simply scanning the entire file to find
    all indirect objects."""

    def __init__(
        self, doc: Union["Document", None] = None, pos: int = 0, offset: int = 0
    ) -> None:
        self.offsets: Dict[int, XRefPos] = {}
        self.trailer: Dict[str, PDFObject] = {}
        # Create a new IndirectObjectParser without a parent document
        # to avoid endless looping
        if doc is not None:
            self._load(IndirectObjectParser(doc.buffer, doc=None, pos=pos), doc)

    def __repr__(self) -> str:
        return "<XRefFallback: offsets=%r>" % (self.offsets.keys())

    def _load(self, parser: IndirectObjectParser, doc: "Document") -> None:
        # Get all the objects
        for m in FIND_INDOBJR.finditer(parser.buffer):
            pos = m.start(0)
            log.debug("Indirect object at %d: %r", pos, m.group(0))
            parser.seek(pos)
            pos, obj = next(parser)
            prev_genno = -1
            if obj.objid in self.offsets:
                prev_genno = self.offsets[obj.objid].genno
                # Apparently this isn't an error, nothing requires you
                # to update the generation number!  (what is it good
                # for anyway then?)  PDF 1.7 section 7.5.6
                # (Incremental Updates): As shown in Figure 3, a file
                # that has been updated several times contains several
                # trailers. Because updates are appended to PDF files,
                # a file may have several copies of an object with the
                # same object identifier (object number and generation
                # number).
                if obj.genno == prev_genno:
                    log.debug(
                        "Duplicate object %d %d at %d: %r",
                        obj.objid,
                        obj.genno,
                        pos,
                        obj.obj,
                    )
            if obj.genno >= prev_genno:
                self.offsets[obj.objid] = XRefPos(None, pos, obj.genno)
            # Expand any object streams right away
            if not isinstance(obj.obj, ContentStream):
                continue
            stream_type = obj.obj.get("Type")
            if stream_type is LITERAL_OBJSTM:
                stream = stream_value(obj.obj)
                try:
                    n = stream["N"]
                except KeyError:
                    log.warning("N is not defined in object stream: %r", stream)
                    n = 0
                parser1 = ObjectParser(stream.buffer, doc)
                objs: List = [obj for _, obj in parser1]
                # FIXME: This is choplist
                n = min(n, len(objs) // 2)
                for index in range(n):
                    objid1 = objs[index * 2]
                    self.offsets[objid1] = XRefPos(obj.objid, index, 0)
                # If we find a cross-reference stream, use it as the trailer
            elif stream_type is LITERAL_XREF:
                # See below re: salvage operation
                self.trailer.update(obj.obj.attrs)
        if self.trailer:
            _update_refs(self.trailer, doc)
            return
        # Get the trailer if we didn't find one.  Maybe there are
        # multiple trailers.  Because this is a salvage operation, we
        # will simply agglomerate them - due to incremental updates
        # the last one should be the most recent, but we can't count
        # on it being complete or correct.
        pos = 0
        while True:
            pos = parser.buffer.find(b"trailer", pos)
            if pos == -1:
                break
            pos += len(b"trailer")
            log.debug("Found possible trailer at %d", pos)
            try:
                _, trailer = next(ObjectParser(parser.buffer, doc, pos))
            except (TypeError, PDFSyntaxError):  # pragma: no cover
                # This actually can't happen because ObjectParser will
                # never throw an exception without strict mode (which
                # we won't turn on when doing fallback parsing)
                continue
            if not isinstance(trailer, dict):
                continue
            log.debug("Trailer: %r", trailer)
            self.trailer.update(trailer)
        if not self.trailer:
            log.warning("b'trailer' not found in document or invalid")

    def __len__(self) -> int:
        return len(self.offsets)

    def __iter__(self) -> Iterator[int]:
        return iter(self.offsets)

    def __getitem__(self, objid: int) -> XRefPos:
        return self.offsets[objid]


class XRefStream(XRef):
    """Cross-reference stream (as of PDF 1.5)"""

    def __init__(
        self, doc: Union["Document", None] = None, pos: int = 0, offset: int = 0
    ) -> None:
        self.offset = offset
        self.data: Union[bytes, None] = None
        self.entlen: Union[int, None] = None
        self.fl1: Union[int, None] = None
        self.fl2: Union[int, None] = None
        self.fl3: Union[int, None] = None
        self.ranges: List[Tuple[int, int]] = []
        # Because an XRefStream's dictionary may contain indirect
        # object references, we create a new IndirectObjectParser
        # here with no document to avoid trying to follow them
        # (and thus creating an infinite loop)
        if doc is not None:
            self._load(IndirectObjectParser(doc.buffer, doc=None, pos=pos), doc)

    def __repr__(self) -> str:
        return "<XRefStream: ranges=%r>" % (self.ranges)

    def _load(self, parser: IndirectObjectParser, doc: "Document") -> None:
        (_, obj) = next(parser)
        stream = obj.obj
        if (
            not isinstance(stream, ContentStream)
            or stream.get("Type") is not LITERAL_XREF
        ):
            raise ValueError(f"Invalid PDF stream spec {stream!r}")
        size = stream["Size"]
        index_array = list_value(stream.get("Index") or [0, size])
        if len(index_array) % 2 != 0:
            raise PDFSyntaxError("Invalid index number")
        for start, end in choplist(2, index_array):
            self.ranges.append((int_value(start), int_value(end)))
        (self.fl1, self.fl2, self.fl3) = stream["W"]
        assert self.fl1 is not None and self.fl2 is not None and self.fl3 is not None
        self.data = stream.buffer
        self.entlen = self.fl1 + self.fl2 + self.fl3
        self.trailer = stream.attrs
        # Update any references in trailer to point to the document
        _update_refs(self.trailer, doc)
        # Dump out objects for debugging
        if log.isEnabledFor(logging.DEBUG):
            for start, nobjs in self.ranges:
                log.debug("objects %d - %d:", start, start + nobjs)
                for index in range(nobjs):
                    offset = self.entlen * index
                    ent = self.data[offset : offset + self.entlen]
                    f1 = nunpack(ent[: self.fl1], 1)
                    f2 = nunpack(ent[self.fl1 : self.fl1 + self.fl2])
                    f3 = nunpack(ent[self.fl1 + self.fl2 :])
                    log.debug("obj %d => %d %d %d", start + index, f1, f2, f3)

    def __iter__(self) -> Iterator[int]:
        for start, nobjs in self.ranges:
            for i in range(nobjs):
                assert self.entlen is not None
                assert self.data is not None
                offset = self.entlen * i
                ent = self.data[offset : offset + self.entlen]
                f1 = nunpack(ent[: self.fl1], 1)
                if f1 == 1 or f1 == 2:
                    yield start + i

    def __len__(self) -> int:
        return sum(nobjs for _, nobjs in self.ranges)

    def __getitem__(self, objid: int) -> XRefPos:
        index = 0
        for start, nobjs in self.ranges:
            if start <= objid and objid < start + nobjs:
                index += objid - start
                break
            else:
                index += nobjs
        else:
            raise KeyError(objid)
        assert self.entlen is not None
        assert self.data is not None
        assert self.fl1 is not None and self.fl2 is not None and self.fl3 is not None
        offset = self.entlen * index
        ent = self.data[offset : offset + self.entlen]
        f1 = nunpack(ent[: self.fl1], 1)
        f2 = nunpack(ent[self.fl1 : self.fl1 + self.fl2])
        f3 = nunpack(ent[self.fl1 + self.fl2 :])
        if f1 == 1:  # not in an object stream
            return XRefPos(None, f2 + self.offset, f3)
        elif f1 == 2:  # in an object stream
            return XRefPos(f2, f3, 0)
        else:
            # this is a free object
            raise KeyError(objid)



import logging
import zlib
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Final,
    Iterable,
    Iterator,
    List,
    Mapping,
    Protocol,
    Tuple,
    Union,
)

from playa.worker import DocumentRef, _deref_document

if TYPE_CHECKING:
    from playa.color import ColorSpace

logger: Final = logging.getLogger(__name__)
PDFObject = Union[
    int,
    float,
    bool,
    "PSLiteral",
    bytes,
    List,
    Dict,
    "ObjRef",
    "PSKeyword",
    "InlineImage",
    "ContentStream",
    None,
]
Point = Tuple[float, float]
Rect = Tuple[float, float, float, float]
Matrix = Tuple[float, float, float, float, float, float]
# Cannot be final because of https://github.com/mypyc/mypyc/issues/1183
MATRIX_IDENTITY: Matrix = (1.0, 0.0, 0.0, 1.0, 0.0, 0.0)


class PSLiteral:
    """A class that represents a PostScript literal.

    Postscript literals are used as identifiers, such as variable
    names, property names and dictionary keys.  Literals are case
    sensitive and denoted by a preceding slash sign (e.g. "/Name").
    They are globally unique objects stored in PSLiteralTable.
    """

    name: str

    def __new__(cls, name: str) -> "PSLiteral":
        if name not in PSLiteralTable:
            PSLiteralTable[name] = object.__new__(cls)
            PSLiteralTable[name].name = name
        return PSLiteralTable[name]

    def __getnewargs__(self) -> Tuple:
        return (self.name,)

    def __repr__(self) -> str:
        return "/%r" % self.name


class PSKeyword:
    """A class that represents a PostScript keyword.

    PostScript keywords are a dozen of predefined words.  Commands and
    directives in PostScript are expressed by keywords.  They are also
    used to denote the content boundaries.  They are globally unique
    objects stored in PSKeywordTable.
    """

    name: bytes

    def __new__(cls, name: bytes) -> "PSKeyword":
        if name not in PSKeywordTable:
            PSKeywordTable[name] = object.__new__(cls)
            PSKeywordTable[name].name = name
        return PSKeywordTable[name]

    def __getnewargs__(self) -> Tuple:
        return (self.name,)

    def __repr__(self) -> str:
        return "/%r" % self.name


# Do not make these generic as they are performance-critical
PSLiteralTable: Final[Dict[str, PSLiteral]] = {}
PSKeywordTable: Final[Dict[bytes, PSKeyword]] = {}

# Compatibility aliases
LIT: Final = PSLiteral
KWD: Final = PSKeyword

# Intern a bunch of important literals
LITERAL_CRYPT: Final = LIT("Crypt")
LITERAL_IMAGE: Final = LIT("Image")
# Abbreviation of Filter names in PDF 4.8.6. "Inline Images"
LITERALS_FLATE_DECODE: Final = (LIT("FlateDecode"), LIT("Fl"))
LITERALS_LZW_DECODE: Final = (LIT("LZWDecode"), LIT("LZW"))
LITERALS_ASCII85_DECODE: Final = (LIT("ASCII85Decode"), LIT("A85"))
LITERALS_ASCIIHEX_DECODE: Final = (LIT("ASCIIHexDecode"), LIT("AHx"))
LITERALS_RUNLENGTH_DECODE: Final = (LIT("RunLengthDecode"), LIT("RL"))
LITERALS_CCITTFAX_DECODE: Final = (LIT("CCITTFaxDecode"), LIT("CCF"))
LITERALS_DCT_DECODE: Final = (LIT("DCTDecode"), LIT("DCT"))
LITERALS_JBIG2_DECODE: Final = (LIT("JBIG2Decode"),)
LITERALS_JPX_DECODE: Final = (LIT("JPXDecode"),)


def name_str(x: bytes) -> str:
    """Get the string representation for a name object.

    According to the PDF 1.7 spec (p.18):

    > Ordinarily, the bytes making up the name are never treated as
    > text to be presented to a human user or to an application
    > external to a conforming reader. However, occasionally the need
    > arises to treat a name object as text... In such situations, the
    > sequence of bytes (after expansion of NUMBER SIGN sequences, if
    > any) should be interpreted according to UTF-8.

    Accordingly, if they *can* be decoded to UTF-8, then they *will*
    be, and if not, we will just decode them as ISO-8859-1 since that
    gives a unique (if possibly nonsensical) value for an 8-bit string.
    """
    try:
        return x.decode("utf-8")
    except UnicodeDecodeError:
        return x.decode("iso-8859-1")


def literal_name(x: Any) -> str:
    if not isinstance(x, PSLiteral):
        raise TypeError(f"Literal required: {x!r}")
    else:
        return x.name


def keyword_name(x: Any) -> str:
    if not isinstance(x, PSKeyword):
        raise TypeError("Keyword required: %r" % x)
    else:
        # PDF keywords are *not* UTF-8 (they aren't ISO-8859-1 either,
        # but this isn't very important, we just want some
        # unique representation of 8-bit characters, as above)
        name = x.name.decode("iso-8859-1")
    return name


class DecipherCallable(Protocol):
    """Fully typed decipher callback, with optional parameter."""

    def __call__(
        self,
        objid: int,
        genno: int,
        data: bytes,
        attrs: Union[Dict[str, Any], None] = None,
    ) -> bytes: ...


class ObjRef:
    def __init__(
        self,
        doc: Union[DocumentRef, None] = None,
        objid: int = 0,
    ) -> None:
        """Reference to a PDF object.

        :param doc: The PDF document.
        :param objid: The object number.
        """
        self.doc = doc
        self.objid = objid

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, ObjRef):
            raise NotImplementedError("Unimplemented comparison with non-ObjRef")
        if self.doc is None and other.doc is None:
            return self.objid == other.objid
        elif self.doc is None or other.doc is None:
            return False
        else:
            selfdoc = _deref_document(self.doc)
            otherdoc = _deref_document(other.doc)
            return selfdoc is otherdoc and self.objid == other.objid

    def __hash__(self) -> int:
        return self.objid

    def __repr__(self) -> str:
        return "<ObjRef:%d>" % (self.objid)

    def resolve(self, default: PDFObject = None) -> PDFObject:
        if self.doc is None:
            return default
        doc = _deref_document(self.doc)
        try:
            return doc[self.objid]
        except KeyError:
            return default


def resolve1(x: PDFObject, default: PDFObject = None) -> PDFObject:
    """Resolves an object.

    If this is an array or dictionary, it may still contains
    some indirect objects inside.
    """
    while isinstance(x, ObjRef):
        x = x.resolve(default=default)
    return x


def resolve_all(x: PDFObject, default: PDFObject = None) -> PDFObject:
    """Resolves all indirect object references inside the given object.

    This creates new copies of any lists or dictionaries, so the
    original object is not modified.  However, it will ultimately
    create circular references if they exist, so beware.
    """

    def resolver(
        x: PDFObject, default: PDFObject, seen: Dict[int, PDFObject]
    ) -> PDFObject:
        if isinstance(x, ObjRef):
            ref = x
            while isinstance(x, ObjRef):
                if x.objid in seen:
                    return seen[x.objid]
                x = x.resolve(default=default)
            seen[ref.objid] = x
        if isinstance(x, list):
            return [resolver(v, default, seen) for v in x]
        elif isinstance(x, dict):
            return {k: resolver(v, default, seen) for k, v in x.items()}
        return x

    return resolver(x, default, {})


def decipher_all(
    decipher: DecipherCallable, objid: int, genno: int, x: PDFObject
) -> PDFObject:
    """Recursively deciphers the given object."""
    if isinstance(x, bytes):
        if len(x) == 0:
            return x
        return decipher(objid, genno, x)
    if isinstance(x, list):
        x = [decipher_all(decipher, objid, genno, v) for v in x]
    elif isinstance(x, dict):
        return {k: decipher_all(decipher, objid, genno, v) for k, v in x.items()}
    return x


def bool_value(x: PDFObject) -> bool:
    x = resolve1(x)
    if not isinstance(x, bool):
        raise TypeError("Boolean required: %r" % (x,))
    return x


def int_value(x: PDFObject) -> int:
    x = resolve1(x)
    if not isinstance(x, int):
        raise TypeError("Integer required: %r" % (x,))
    return x


def float_value(x: PDFObject) -> float:
    x = resolve1(x)
    if not isinstance(x, float):
        raise TypeError("Float required: %r" % (x,))
    return x


def num_value(x: PDFObject) -> float:
    x = resolve1(x)
    if not isinstance(x, (int, float)):
        raise TypeError("Int or Float required: %r" % x)
    return x


def uint_value(x: PDFObject, n_bits: int) -> int:
    """Resolve number and interpret it as a two's-complement unsigned number"""
    xi = int_value(x)
    if xi > 0:
        return xi
    else:
        return xi + (1 << n_bits)


def str_value(x: PDFObject) -> bytes:
    x = resolve1(x)
    if not isinstance(x, bytes):
        raise TypeError("String required: %r" % x)
    return x


def list_value(x: PDFObject) -> Union[List[Any], Tuple[Any, ...]]:
    x = resolve1(x)
    if not isinstance(x, (list, tuple)):
        raise TypeError("List required: %r" % x)
    return x


def dict_value(x: PDFObject) -> Dict[Any, Any]:
    x = resolve1(x)
    if not isinstance(x, dict):
        raise TypeError("Dict required: %r" % x)
    return x


def stream_value(x: PDFObject) -> "ContentStream":
    x = resolve1(x)
    if not isinstance(x, ContentStream):
        raise TypeError("ContentStream required: %r" % x)
    return x


def point_value(o: PDFObject) -> Point:
    try:
        lp = list_value(o)
        if len(lp) != 2:
            raise ValueError("Point must have 2 elements")
        x = num_value(lp[0])
        y = num_value(lp[1])
        return x, y
    except ValueError:
        raise ValueError("Could not parse point %r" % (o,))
    except TypeError:
        raise TypeError("Point contains non-numeric values")


def rect_value(o: PDFObject) -> Rect:
    try:
        lr = list_value(o)
        if len(lr) != 4:
            raise ValueError("Rect must have 4 elements")
        x0 = num_value(lr[0])
        y0 = num_value(lr[1])
        x1 = num_value(lr[2])
        y1 = num_value(lr[3])
        return x0, y0, x1, y1
    except ValueError:
        raise ValueError("Could not parse rectangle %r" % (o,))
    except TypeError:
        raise TypeError("Rectangle contains non-numeric values")


def matrix_value(o: PDFObject) -> Matrix:
    try:
        lm = list_value(o)
        if len(lm) != 6:
            raise ValueError("Matrix must have 6 elements")
        a = num_value(lm[0])
        b = num_value(lm[1])
        c = num_value(lm[2])
        d = num_value(lm[3])
        e = num_value(lm[4])
        f = num_value(lm[5])
        return a, b, c, d, e, f
    except ValueError:
        raise ValueError("Could not parse matrix %r" % (o,))
    except TypeError:
        raise TypeError("Matrix contains non-numeric values")


def decompress_corrupted(data: bytes, bufsiz: int = 4096) -> bytes:
    """Decompress (possibly with data loss) a corrupted FlateDecode stream."""
    d = zlib.decompressobj()
    size = len(data)
    result_str = b""
    pos = end = 0
    try:
        while pos < size:
            # Skip the CRC checksum unless it's the only thing left
            end = min(size - 3, pos + bufsiz)
            if end == pos:
                end = size
            result_str += d.decompress(data[pos:end])
            pos = end
            logger.debug(
                "decompress_corrupted: %d bytes in, %d bytes out", pos, len(result_str)
            )
    except zlib.error as e:
        # Let the error propagates if we're not yet in the CRC checksum
        if pos != size - 3:
            logger.warning(
                "Data loss in decompress_corrupted: %s: bytes %d:%d", e, pos, end
            )
    return result_str


class ContentStream(Mapping[str, PDFObject]):
    _data: Union[bytes, None] = None
    _colorspace: Union["ColorSpace", None] = None
    objid: Union[int, None] = None
    genno: Union[int, None] = None

    def __init__(
        self,
        attrs: Union[Dict[str, Any], None] = None,
        rawdata: bytes = b"",
        decipher: Union[DecipherCallable, None] = None,
    ) -> None:
        if attrs is None:
            attrs = {}
        self.attrs = attrs
        self.rawdata = rawdata
        self.decipher = decipher

    def __repr__(self) -> str:
        if self._data is None:
            return "<ContentStream(%r): raw=%d, %r>" % (
                self.objid,
                len(self.rawdata),
                self.attrs,
            )
        else:
            return "<ContentStream(%r): len=%d, %r>" % (
                self.objid,
                len(self._data),
                self.attrs,
            )

    def __contains__(self, name: object) -> bool:
        return name in self.attrs

    def __getitem__(self, name: str) -> Any:
        return self.attrs[name]

    def __iter__(self) -> Iterator[str]:
        return iter(self.attrs)

    def __len__(self) -> int:
        return len(self.attrs)

    def get_any(self, names: Iterable[str], default: PDFObject = None) -> PDFObject:
        for name in names:
            if name in self.attrs:
                return self.attrs[name]
        return default

    @property
    def filters(self) -> List[PSLiteral]:
        filters = resolve1(self.get_any(("F", "Filter")))
        if not filters:
            return []
        if not isinstance(filters, list):
            filters = [filters]
        return [f for f in filters if isinstance(f, PSLiteral)]

    def get_filters(self) -> List[Tuple[PSLiteral, Dict[str, PDFObject]]]:
        filters = self.filters
        params = resolve1(self.get_any(("DP", "DecodeParms", "FDecodeParms")))
        if not params:
            params = {}
        if not isinstance(params, list):
            params = [params] * len(filters)
        resolved_params = []
        for p in params:
            rp = resolve1(p)
            if isinstance(rp, dict):
                resolved_params.append(rp)
            else:
                resolved_params.append({})
        return list(zip(filters, resolved_params))

    def decode(self, strict: bool = False) -> bytes:
        data = self.rawdata
        if self.decipher:
            # Handle encryption
            assert self.objid is not None
            assert self.genno is not None
            data = self.decipher(self.objid, self.genno, data, self.attrs)
        filters = self.get_filters()
        if not filters:
            self._data = data
            return data
        for f, params in filters:
            if f in LITERALS_FLATE_DECODE:
                # will get errors if the document is encrypted.
                try:
                    data = zlib.decompress(data)
                except zlib.error as e:
                    if strict:
                        error_msg = f"Invalid zlib bytes: {e!r}, {data!r}"
                        raise ValueError(error_msg)
                    else:
                        logger.warning("%s: %r", e, self)
                    data = decompress_corrupted(data)

            elif f in LITERALS_LZW_DECODE:
                from playa.lzw import lzwdecode

                data = lzwdecode(data)
            elif f in LITERALS_ASCII85_DECODE:
                from playa.ascii85 import ascii85decode

                data = ascii85decode(data)
            elif f in LITERALS_ASCIIHEX_DECODE:
                from playa.ascii85 import asciihexdecode

                data = asciihexdecode(data)
            elif f in LITERALS_RUNLENGTH_DECODE:
                from playa.runlength import rldecode

                data = rldecode(data)
            elif f in LITERALS_CCITTFAX_DECODE:
                from playa.ccitt import ccittfaxdecode

                data = ccittfaxdecode(data, params)
            elif f in LITERALS_DCT_DECODE:
                # This is probably a JPG stream
                # it does not need to be decoded twice.
                # Just return the stream to the user.
                pass
            elif f in LITERALS_JBIG2_DECODE or f in LITERALS_JPX_DECODE:
                pass
            elif f == LITERAL_CRYPT:
                # not yet..
                raise NotImplementedError("/Crypt filter is unsupported")
            else:
                raise NotImplementedError("Unsupported filter: %r" % f)
            # apply predictors
            if params and "Predictor" in params:
                pred = int_value(params["Predictor"])
                if pred == 1:
                    # no predictor
                    pass
                elif pred == 2:
                    # TIFF predictor 2
                    from playa.utils import apply_tiff_predictor

                    colors = int_value(params.get("Colors", 1))
                    columns = int_value(params.get("Columns", 1))
                    raw_bits_per_component = params.get("BitsPerComponent", 8)
                    bitspercomponent = int_value(raw_bits_per_component)
                    data = apply_tiff_predictor(
                        colors,
                        columns,
                        bitspercomponent,
                        data,
                    )
                elif pred >= 10:
                    # PNG predictor
                    from playa.utils import apply_png_predictor

                    colors = int_value(params.get("Colors", 1))
                    columns = int_value(params.get("Columns", 1))
                    raw_bits_per_component = params.get("BitsPerComponent", 8)
                    bitspercomponent = int_value(raw_bits_per_component)
                    data = apply_png_predictor(
                        pred,
                        colors,
                        columns,
                        bitspercomponent,
                        data,
                    )
                else:
                    error_msg = "Unsupported predictor: %r" % pred
                    raise NotImplementedError(error_msg)
        self._data = data
        return data

    @property
    def bits(self) -> int:
        """Bits per component for an image stream.

        Default is 1."""
        return int_value(self.get_any(("BPC", "BitsPerComponent"), 1))

    @property
    def width(self) -> int:
        """Width in pixels of an image stream.

        It may be the case that a stream has no inherent width, in
        which case the default width is 1.
        """
        return int_value(self.get_any(("W", "Width"), 1))

    @property
    def height(self) -> int:
        """Height in pixels for an image stream.

        It may be the case that a stream has no inherent height, in
        which case the default height is 1."""
        return int_value(self.get_any(("H", "Height"), 1))

    @property
    def colorspace(self) -> "ColorSpace":
        """Colorspace for an image stream.

        Default is DeviceGray (1 component).

        Raises: ValueError if the colorspace is invalid, or
                unfortunately also in the case where it is a named
                resource in the containing page (or Form XObject, in
                the case of an inline image) and the stream is
                accessed from outside an interpreter for that
                page/object.
        """
        from playa.color import get_colorspace, LITERAL_DEVICE_GRAY

        if self._colorspace is not None:
            return self._colorspace
        spec = resolve1(self.get_any(("CS", "ColorSpace"), LITERAL_DEVICE_GRAY))
        cs = get_colorspace(spec)
        if cs is None:
            raise ValueError("Unknown or undefined colour space: %r" % (spec,))
        self._colorspace: "ColorSpace" = cs
        return self._colorspace

    @colorspace.setter
    def colorspace(self, cs: "ColorSpace") -> None:
        self._colorspace = cs

    @property
    def buffer(self) -> bytes:
        """The decoded contents of the stream."""
        return self.decode()


class InlineImage(ContentStream):
    """Specific class for inline images so the interpreter can
    recognize them (they are otherwise the same thing as content
    streams)."""



"""
Benchmark latency of opening PDFs (possibly other things soon).
"""

import logging
import time
from pathlib import Path

import playa

LOG = logging.getLogger("benchmark-convert")
# Use a standard benchmark set to make version comparisons possible
CONTRIB = Path(__file__).parent.parent / "samples" / "contrib"
PDFS = [
    "2023-04-06-ODJ et Résolutions-séance xtra 6 avril 2023.pdf",
    "2023-06-20-PV.pdf",
    "PSC_Station.pdf",
    "Rgl-1314-2021-DM-Derogations-mineures.pdf",
    "issue-886-xref-stream-widths.pdf",
    "issue-146-broken-xref-and-streams.pdf",
    "issue-1008-inline-ascii85.pdf",
    "evil-pi-to-100000-digits.pdf",
]


def benchmark_latency(*, ncpus: int = 1) -> None:
    cat_time = fonts_time = open_time = page0_time = 0.0
    nfiles = 0
    niter = 5
    for idx in range(niter + 1):
        for name in PDFS:
            path = CONTRIB / name
            start = time.time()
            pdf = playa.open(path, max_workers=ncpus)
            if idx != 0:
                open_time += time.time() - start
                nfiles += 1
            _ = pdf.catalog
            if idx != 0:
                cat_time += time.time() - start
            try:
                page = next(iter(pdf.pages))
            except StopIteration:
                continue
            if idx != 0:
                page0_time += time.time() - start
            _ = page.fonts
            if idx != 0:
                fonts_time += time.time() - start

    print("PLAYA (%d cpu%s)" % (ncpus, "s" if ncpus > 1 else ""))
    print("Open took %.3f ms / file" % (open_time / nfiles * 1000,))
    print("Catalog took %.3f ms / file" % (cat_time / nfiles * 1000,))
    print("Page 0 took %.3f ms / file" % (page0_time / nfiles * 1000,))
    print("Page 0 Fonts took %.3f ms / file" % (fonts_time / nfiles * 1000,))


if __name__ == "__main__":
    logging.basicConfig(level=logging.ERROR)
    # Benchmark latency without parallelism
    benchmark_latency(ncpus=1)
    # Ensure that parallelism doesn't add latency
    benchmark_latency(ncpus=2)



"""
Test the ContentObject API for pages.
"""

from pathlib import Path
from typing import Iterable

import playa
import pytest
from playa.color import PREDEFINED_COLORSPACE, Color
from playa.content import ContentObject
from playa.exceptions import PDFEncryptionError
from playa.utils import get_bound
from playa.image import get_one_image

from .data import ALLPDFS, CONTRIB, PASSWORDS, TESTDIR, XFAILS


@pytest.mark.skipif(not CONTRIB.exists(), reason="contrib samples not present")
def test_content_objects():
    """Ensure that we can produce all the basic content objects."""
    with playa.open(CONTRIB / "2023-06-20-PV.pdf", space="page") as pdf:
        page = pdf.pages[0]
        img = next(page.images)
        assert img.colorspace.name == "ICCBased"
        assert img.colorspace.ncomponents == 3
        ibbox = [round(x) for x in img.bbox]
        assert ibbox == [254, 899, 358, 973]
        mcs_bbox = img.mcs.props["BBox"]
        # Not quite the same, for Reasons!
        assert mcs_bbox == [254.25, 895.5023, 360.09, 972.6]
        for obj in page.paths:
            assert obj.object_type == "path"
            # We cannot directly iterate over path segments becuase
            # they aren't ContentObjects
            assert len(obj) == 0
        rect = next(obj for obj in page.paths)
        ibbox = [round(x) for x in rect.bbox]
        assert ibbox == [85, 669, 211, 670]
        boxes = []
        texts = []
        for obj in page.texts:
            assert obj.object_type == "text"
            ibbox = [round(x) for x in obj.bbox]
            boxes.append(ibbox)
            texts.append(obj.chars)
            assert len(obj) == sum(1 for glyph in obj)
        # Now there are ... a lot of text objects
        assert boxes[0] == [358, 896, 360, 909]
        assert boxes[-1] == [99, 79, 102, 94]
        assert len(boxes) == 204


@pytest.mark.parametrize("path", ALLPDFS, ids=str)
def test_open_lazy(path: Path) -> None:
    """Open all the documents"""
    if path.name in XFAILS:
        pytest.xfail("Intentionally corrupt file: %s" % path.name)
    passwords = PASSWORDS.get(path.name, [""])
    for password in passwords:
        beach = []
        try:
            with playa.open(path, password=password) as doc:
                for page in doc.pages:
                    for obj in page:
                        try:
                            beach.append((obj.object_type, obj.bbox))
                        except ValueError as e:
                            if "not enough values" in str(e):
                                continue
                            raise e
        except PDFEncryptionError:
            pytest.skip("cryptography package not installed")


def test_uncoloured_tiling() -> None:
    """Verify that we handle uncoloured tiling patterns correctly."""
    with playa.open(TESTDIR / "uncoloured-tiling-pattern.pdf") as pdf:
        paths = pdf.pages[0].paths
        path = next(paths)
        assert path.gstate.ncs == PREDEFINED_COLORSPACE["DeviceRGB"]
        assert path.gstate.ncolor == Color((1.0, 1.0, 0.0), None)
        path = next(paths)
        assert path.gstate.ncolor == Color((0.77, 0.2, 0.0), "P1")
        path = next(paths)
        assert path.gstate.ncolor == Color((0.2, 0.8, 0.4), "P1")
        path = next(paths)
        assert path.gstate.ncolor == Color((0.3, 0.7, 1.0), "P1")
        path = next(paths)
        assert path.gstate.ncolor == Color((0.5, 0.2, 1.0), "P1")


@pytest.mark.skipif(not CONTRIB.exists(), reason="contrib samples not present")
def test_rotated_glyphs() -> None:
    """Verify that we (unlike pdfminer) properly calculate the bbox
    for rotated text."""
    with playa.open(CONTRIB / "issue_495_pdfobjref.pdf") as pdf:
        chars = []
        for text in pdf.pages[0].texts:
            for glyph in text:
                if 1 not in text.line_matrix:
                    if glyph.text is not None:
                        chars.append(glyph.text)
                    x0, y0, x1, y1 = glyph.bbox
                    width = x1 - x0
                    assert width > 6
        assert "".join(chars) == "R18,00"


def test_rotated_text_objects() -> None:
    """Verify specializations of bbox for text."""
    with playa.open(TESTDIR / "rotated.pdf") as pdf:
        # Ensure that the text bbox is the same as the bounds of the
        # glyph bboxes (this will also ensure no side effects)
        for text in pdf.pages[0].texts:
            bbox = text.bbox
            points = []
            for glyph in text:
                x0, y0, x1, y1 = glyph.bbox
                print(glyph.text, ":", glyph.bbox)
                points.append((x0, y0))
                points.append((x1, y1))
            assert bbox == pytest.approx(get_bound(points))


def test_text_displacement() -> None:
    with playa.open(TESTDIR / "text_displacement.pdf") as pdf:
        x, y = (100.0, 200.0)
        for text in pdf.pages[0].texts:
            cx, cy = text.origin
            assert cx == pytest.approx(x)
            assert cy == pytest.approx(y)
            dx, dy = text.displacement
            print(text.chars, cx, cy, dx, dy)
            x += dx
            y += dy
        x, y = (100.0, 200.0)
        for glyph in pdf.pages[0].glyphs:
            cx, cy = glyph.origin
            assert cx == pytest.approx(x)
            assert cy == pytest.approx(y)
            dx, dy = glyph.displacement
            print(glyph.text, cx, cy, dx, dy)
            x += dx
            y += dy


TEXTOBJS = [
    {
        "chars": "foo",
        "bbox": [0.0, -4.968, 33.36, 17.232],
    },
    {
        "chars": "A",
        "bbox": [50.0, 95.032, 66.00800000000001, 117.232],
    },
    {
        "chars": "B",
        "bbox": [99.012, 142.548, 123.024, 175.848],
    },
    {
        "chars": "C",
        "bbox": [184.536, 213.822, 223.524, 263.772],
    },
    {
        "chars": "D",
        "bbox": [223.524, 213.822, 262.51200000000006, 263.772],
    },
    {
        "chars": "BAR",
        "bbox": [262.51200000000006, 213.822, 373.53600000000006, 263.772],
    },
    {
        "chars": "FOO",
        "bbox": [0.0, -11.178, 117.01799999999999, 38.772],
    },
    {
        "chars": "Hello World",
        "bbox": [0.0, 370.032, 124.00800000000004, 392.23199999999997],
    },
]


def test_operators_in_text() -> None:
    """Verify that other operators are properly ordered in text objects."""
    # Verify consistent handling of graphics state and text state
    with playa.open(
        TESTDIR / "graphics_state_in_text_object.pdf", space="default"
    ) as pdf:
        page = pdf.pages[0]
        for text, obj in zip(page.texts, TEXTOBJS):
            assert text.chars == obj["chars"]
            assert text.bbox == pytest.approx(obj["bbox"])
            if text.chars == "B":
                assert text.ctm[0] == 1.5
                assert text.gstate.ncs.name == "DeviceRGB"
                assert text.gstate.ncolor.values == (0.75, 0.25, 0.25)

    # Also verify that calling TJ with no actual text still does something
    with playa.open(TESTDIR / "text_side_effects.pdf") as pdf:
        boxes = [[g.bbox for g in t] for t in pdf.pages[0].texts]
        # there was a -5000 that moved it right
        assert boxes[0][0][0] >= 170
        # and a -1000 that moved it right some more
        assert boxes[1][0][0] >= 210
    # Also verify that we get the right ActualText and MCID
    with playa.open(TESTDIR / "actualtext.pdf") as pdf:
        for t in pdf.pages[0].texts:
            if t.mcs and "ActualText" in t.mcs.props:
                assert isinstance(t.mcs.props["ActualText"], bytes)
                assert t.mcs.props["ActualText"].decode("utf-16") == "x̌"
            assert t.mcid == 0


def test_broken_xobjects() -> None:
    """Verify that we tolerate missing attributes on XObjects."""
    with playa.open(TESTDIR / "broken_xobjects.pdf") as doc:
        page = doc.pages[0]
        for img in page.images:
            assert img.srcsize == (1, 1)
            assert img.bbox == (25.0, 154.0, 237.0, 275.0)
        for xobj in page.xobjects:
            assert xobj.bbox == page.cropbox


@pytest.mark.skipif(not CONTRIB.exists(), reason="contrib samples not present")
def test_glyph_bboxes() -> None:
    """Verify that we don't think all fonts are 1000 units high."""
    with playa.open(CONTRIB / "issue-79" / "test.pdf") as doc:
        page = doc.pages[0]
        texts = page.texts
        t = next(texts)
        assert t.bbox is not None
        _, zh_y0, _, zh_y1 = t.bbox
        t = next(texts)
        assert t.bbox is not None
        _, en_y0, _, en_y1 = t.bbox
        assert en_y0 <= zh_y0
        assert en_y1 >= zh_y1


@pytest.mark.parametrize("name", ["rotated.pdf", "character_spacing.pdf"])
def test_glyph_properties(name: str) -> None:
    """Check that the newfangled glyph properties do what they should."""
    # Ensure that origin and displacement reflect actual positions
    for space in "default", "page", "screen":
        with playa.open(TESTDIR / name, space=space) as pdf:
            for text in pdf.pages[0].texts:
                next_origin = text.origin
                for glyph in text:
                    assert glyph.origin == pytest.approx(next_origin)
                    gx, gy = glyph.origin
                    dx, dy = glyph.displacement
                    next_origin = (gx + dx, gy + dy)
                    assert glyph.size > 0
                assert text.size > 0


@pytest.mark.skipif(not CONTRIB.exists(), reason="contrib samples not present")
def test_jbig2(tmp_path) -> None:
    """Verify that we can extract JBIG2 images."""
    with playa.open(CONTRIB / "pdf-with-jbig2.pdf") as pdf:
        img = next(pdf.pages[0].images)
        hyppath = tmp_path / "XIPLAYER0"
        hyppath = get_one_image(img.stream, hyppath)
        assert hyppath.suffix == ".jb2"
        refdata = (CONTRIB / "XIPLAYER0.jb2").read_bytes()
        hypdata = hyppath.read_bytes()
        assert refdata == hypdata


@pytest.mark.skipif(not CONTRIB.exists(), reason="contrib samples not present")
def test_indexed_images(tmp_path) -> None:
    """Verify that we can extract (at least some) Indexed images
    correctly."""
    with playa.open(CONTRIB / "issue-1062-filters.pdf") as pdf:
        (img,) = pdf.pages[0].images
        outpath = tmp_path / "page1"
        outpath = get_one_image(img.stream, outpath)
        assert outpath.suffix == ".ppm"
        refpath = CONTRIB / "page1-0-00005.ppm"
        hyp = outpath.read_bytes()
        ref = refpath.read_bytes()
        # Testing equality is ABSURDLY STUPIDLY SLOW when it fails
        assert len(hyp) == len(ref)
    with playa.open(CONTRIB / "inline-indexed-images.pdf") as pdf:
        imgs = list(pdf.pages[0].images)
        assert len(imgs) == 7


@pytest.mark.skipif(not CONTRIB.exists(), reason="contrib samples not present")
def test_ccitt(tmp_path) -> None:
    """Verify that we can extract CCITT compressed images correctly."""
    for name in "ccitt-default-k.pdf", "ccitt_EndOfBlock_false.pdf":
        with playa.open(CONTRIB / name) as pdf:
            for img in pdf.pages[0].images:
                get_one_image(img.stream, tmp_path / img.xobjid)
                with open(CONTRIB / "ccitt" / f"{img.xobjid}.pbm", "rb") as fh:
                    ref = fh.read()
                with open(tmp_path / f"{img.xobjid}.pbm", "rb") as fh:
                    hyp = fh.read()
                assert ref == hyp


def test_finalize() -> None:
    """At least minimally verify that finalize() does something useful."""
    with playa.open(TESTDIR / "graphics_state_in_text_object.pdf") as pdf:

        def compare_objects(page: playa.Page, objects: Iterable[ContentObject]):
            for ref, hyp in zip(page, objects):
                for ref_child, hyp_child in zip(ref, hyp):
                    assert ref_child == hyp_child
                # Do this after as we need internal properties
                # (_next_glyph_offset) to exist
                assert ref == hyp

        # First demonstrate "DO NOT do this" from README.md
        with pytest.raises(AssertionError):
            compare_objects(pdf.pages[0], list(pdf.pages[0]))
        # All clear!
        compare_objects(pdf.pages[0], [obj.finalize() for obj in pdf.pages[0]])
